
"""
    genonets_filters
    ~~~~~~~~~~~~~~~~

    Contains filters used throughout the package.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

class WriterFilter :
	@staticmethod
	def gmlAttribsToIgnore(level) :
		if level == "network" :
			attrs = [
				"Evolvability_targets",
				"SqrEpi_list",
				"diameterPath_list",
				"Squares_list"
			]
		elif level == "vertex" :
			attrs = [
				"label",					
				"pathsToSummit",
				"VtxToSqrs"
			]

		return attrs

	@staticmethod
	def netAttribsToIgnore() :
		return	[
					"name",
					"SqrEpi_list",
					"diameterPath_list",
					"Squares_list",
					"Summit"
				]

	@staticmethod
	def seqAttribsToIgnore() :
		return	[
					"sequences",
					"label",
					"escores",
					"pathsToSummit",
					"VtxToSqrs"
				]




"""
    genonets_constants
    ~~~~~~~~~~~~~~~~~~

    Defines constans used throughout the package.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

from genonets_utils import Utils


class GenonetsConstants :
	ALL = 0


class SupportedAlphabet :
	binary = ['0', '1']
	rna = ['A', 'U', 'C', 'G']
	dna = ['A', 'T', 'C', 'G']
	protein = ['A', 'R', 'N', 'D', 'C', 'E', 'Q', 'G', 'H', 'I',  \
			   'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V']

	typeToList = {"RNA": rna, "DNA": dna, "Binary": binary, "Protein": protein}

	@staticmethod
	def getAlphabet(alphabetType) :
		return SupportedAlphabet.typeToList[alphabetType]


class EpistasisConstants :
	# Constants for classes of epistasis
	NO_EPISTASIS = 0
	MAGNITUDE = 1
	SIGN = 2
	RECIPROCAL_SIGN = 3

	# Dictionary to map the classes to string descriptions
	epiToDesc = { MAGNITUDE : "Magnitude epistasis", \
				  SIGN : "Sign epistasis", \
				  RECIPROCAL_SIGN : "Reciprocal sign epistasis", \
				  NO_EPISTASIS : "No epistasis" }

	@staticmethod
	def getDesciption(epiType) :
		try :
			return EpistasisConstants.epiToDesc[epiType]
		except KeyError :
			return ""


class AnalysisConstants :
	# Constants representing analysis types
	ALL = 0
	LANDSCAPE = 1	# FIXME: Is this grouping really necessary? ...
	PEAKS = 2
	PATHS = 3
	EPISTASIS = 4
	ROBUSTNESS = 5
	EVOLVABILITY = 6
	ACCESSIBILITY = 7
	NEIGHBOR_ABUNDANCE = 8
	PHENOTYPIC_DIVERSITY = 9
	STRUCTURE = 10
	OVERLAP = 11

	# Dictionary to map constants to string descriptors
	# TODO: Look into whether 'ALL' and 'LANDSCAPE' should be included
	#		in the following dict ...
	analysisToDesc = { PEAKS : "Peaks",
					   PATHS : "Paths",
					   EPISTASIS : "Epistasis", 
					   ROBUSTNESS : "Robustness",
					   EVOLVABILITY : "Evolvability",
					   ACCESSIBILITY : "Accessibility",
					   NEIGHBOR_ABUNDANCE : "Neighbor Abundance",
					   PHENOTYPIC_DIVERSITY : "Diversity Index",
					   STRUCTURE : "Structure",
					   OVERLAP : "Overlap" }

	@staticmethod
	def getAnalysisTypes() :
		return Utils.reverseDict(AnalysisConstants.analysisToDesc)


class ErrorCodes :
	# Generic unknown error
	UNKNOWN_ERROR = 5000

	# Parsing errors
	UNKNOWN_PARSING_ERROR = 400
	INCONSISTENT_HEADER = 500	
	MISSING_VALUE = 501
	BAD_SCORE_FORMAT = 502
	BAD_DELTA_FORMAT = 503
	INCONSISTENT_SEQ_LEN = 504
	ALPHABET_TYPE_MISMATCH = 505
	NO_USEABLE_SCORES = 506

	# File I/O errors
	CANNOT_WRITE_TO_FILE = 550
	CANNOT_CREATE_DIRECTORY = 551

	# Analysis errors
	NOT_ENOUGH_REPS_OLAP = 700

	# Error code to description dict
	errCodeToDesc = {
		UNKNOWN_ERROR : "Something went wrong while performing analyses",
		UNKNOWN_PARSING_ERROR : "Something went wrong during input file parsing",
		INCONSISTENT_HEADER : "Input file parsing error - Column headers are not consistent with specification",
		MISSING_VALUE : "Input file parsing error - Missing value encountered in input file",
		BAD_SCORE_FORMAT : "Input file parsing error - Score value is not in the supported format",
		BAD_DELTA_FORMAT : "Input file parsing error - Delta value is not in the supported format",
		INCONSISTENT_SEQ_LEN : "Input file parsing error - Inconsistent genotype length encountered. All genotypes must be of equal length",
		ALPHABET_TYPE_MISMATCH : "Input file parsing error - Genotype consists of at least one letter not in the selected alphabet type",
		NO_USEABLE_SCORES : "Input file parsing error - No genotypes found with score values greater than",
		NOT_ENOUGH_REPS_OLAP: "Analysis error - Overlap computation can only be performed if there are least two phenotypes that have genotypes with Score values greater than Tau",
		CANNOT_WRITE_TO_FILE: "Could not write to file",
		CANNOT_CREATE_DIRECTORY: "Error while trying to create directory"
	}

	@staticmethod
	def getErrDescription(errCode) :
		try :
			return ErrorCodes.errCodeToDesc[errCode]
		except :
			return ""



#!/usr/bin/env python

"""
    genonets_exmpl_parallel
    ~~~~~~~~~~~~~~~~~~~~~~~

    Demonstrates the steps required to create genotype networks, perform analyses, 
    and write results to files using Genonets with multiprocessing enabled.

	Use the following command to run the script:
	'python genonets_exmpl_parallel.py DNA true data/genonets_sample_input.txt 0.35 results_parallel'

	Output files will be generated in 'results_parallel/'

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

from genonets import cmdl_handler					# For parsing command line arguments
from genonets import genonets_interface as gn_if	# Interface to get the Genonets object

def process(args) :
	# Create the Genonets object. This will load the input file into
	# memory.
	gn = gn_if.Genonets(args)

	# Use 'gn' to create genotype networks for all genotype sets.
	gn.create(parallel=True)

	# Perform all available analyses on all genotype networks.
	gn.analyze(parallel=True)

	# Write all genotype networks to files in GML format. For a genotype network
	# with two or more components, two files are generated: One corresponds to the
	# entire network with all components, and the other corresponds to the dominant
	# component only.
	gn.save()

	# Save all genotype network level measures to 'Genotype_set_measures.txt'.
	gn.saveNetResults()

	# Save all genotype level measures to '<genotypeSetName>_genotype_measures.txt'
	# files.  One file per genotype set is generated.
	gn.saveGenotypeResults()


if __name__ == "__main__" :
	# Parse the command line arguments using the Genonets command line handler, and
	# pass the list of arguments to 'process()'.
	process(cmdl_handler.CmdParser().getArgs())

	# Print message to indicate processing is done.
	print("Done.\n")




"""
    genonets_reader
    ~~~~~~~~~~~~~~~

    Parser for the genonets input file format.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import csv

from genonets_exceptions import GenonetsError
from genonets_constants import ErrorCodes, SupportedAlphabet


class InReader :
	# Required input file column headers
	COL_HEADERS = ["Genotypeset", "Delta", "Genotype", "Score"]

	# Uses the given DictReader to construct a dictionary from file with,
	# key=Genotypeset, value=dict{key=sequence, value=score}
	# Returns the dictionary
	# TODO: This function can be extend to also return a reverse dictionary
	#		with key=score, value=sequence ...
	@staticmethod
	def buildDataDicts(inFilePath, tau, alphabetType) :
		# Dicts to be returned
		dataDict = {}
		deltaDict = {}
		sequences = []

		# Genotype length to be determined
		seqLength = 0

		# Get handles to the input file and a DictReader for the file
		reader, inFile = InReader.getDictReaderForFile(inFilePath)

		# Check if all the required column headers are available in the file
		if not InReader.reqHdrsArePresent(reader.fieldnames) :
			inFile.close()

			print("Error: " + 
				ErrorCodes.getErrDescription(ErrorCodes.INCONSISTENT_HEADER))

			raise GenonetsError(ErrorCodes.INCONSISTENT_HEADER)

		# For each data row in the file
		for row in reader :
			# Check for missing values in this row
			if any(row[col] in (None, "") for col in row.keys()) :
				inFile.close()

				lineNum = str(int(reader.line_num))

				print("Error: " + 
					ErrorCodes.getErrDescription(ErrorCodes.MISSING_VALUE) +
					": Line No. " + lineNum)

				raise GenonetsError(ErrorCodes.MISSING_VALUE, 
					"Line No. " + lineNum)

			# Get the fitness score
			try :
				score = float(row["Score"])
			except :
				inFile.close()

				lineNum = str(int(reader.line_num))

				print("Error: " + 
					ErrorCodes.getErrDescription(ErrorCodes.BAD_SCORE_FORMAT) +
					": Line No. " + lineNum)

				raise GenonetsError(ErrorCodes.BAD_SCORE_FORMAT,
						"Line No. " + lineNum)

			# If the score for this sequence is greater than or equal to
			# the given threshold,
			if score >= tau :
				# If the current genotypeset has not already been added
				if row["Genotypeset"] not in dataDict :
					# Initialize dict for the genotypeset
					dataDict[row["Genotypeset"]] = {}

					# Get the delta value
					try :
						delta = float(row["Delta"])
					except :
						inFile.close()

						lineNum = str(int(reader.line_num))

						print("Error: " + 
							ErrorCodes.getErrDescription(ErrorCodes.BAD_DELTA_FORMAT) +
							": Line No. " + lineNum)

						raise GenonetsError(
							ErrorCodes.BAD_DELTA_FORMAT,
							"Line No. " + lineNum)

					deltaDict[row["Genotypeset"]] = delta

				# Get the genotype sequence
				sequence = row["Genotype"]

				# If sequence length has not been initialized yet, i.e., this is
				# the first row,
				if seqLength == 0 :
					# Set length of the current genotype as the sequence length
					# for the entire dataset
					seqLength = len(sequence)

				try :
					InReader.verifyGenotype(sequence, seqLength, alphabetType, 
						str(int(reader.line_num)))
				except Exception as e :
					inFile.close()
					raise e

				# Add sequence as key and score as value to the current
				# genotypeset
				dataDict[row["Genotypeset"]][sequence] = score

				# If the sequence has not already been read for any other
				# genotypeset,
				if sequence not in sequences :
					# Add it to the list of unique sequences found in the
					# input file
					sequences.append(sequence)

		inFile.close()

		# If no genotypes were found with score >= tau,
		if not dataDict :
			print("Error: " + 
				ErrorCodes.getErrDescription(ErrorCodes.NO_USEABLE_SCORES)
				+ ": Tau=" + str(tau))

			raise GenonetsError(
					ErrorCodes.NO_USEABLE_SCORES,
					"Tau=" + str(tau))

		# Dictionary: Key=Sequence, Value=[Genotypesets]. Reverse dictionary
		# that is used in functions like evolvability.
		seqToRepDict = InReader.getSeqToRepDict(sequences, dataDict)

		return dataDict, deltaDict, seqToRepDict, seqLength

	# Builds Dictionary: Key=Sequence, Value=[Genotypesets in which this sequence
	# exists]
	@staticmethod
	def getSeqToRepDict(sequences, dataDict) :
		seqToRepDict = {}

		# For each sequence in the given list of sequences,
		for seq in sequences :
			# Initialize the list genotypesets
			seqToRepDict[seq] = []

			# For each genotypeset in the given data dictionary,
			for rep in dataDict.keys() :
				# If the sequence exists in the genotypeset,
				if seq in dataDict[rep] :
					# Append the genotypeset name to the list of values
					# for this sequence
					seqToRepDict[seq].append(rep)

		return seqToRepDict

	# If any of the required column headers are not in the file, returns
	# False. Otherwise return True.
	@staticmethod
	def reqHdrsArePresent(colNames) :
		colsInFile = [col for col in colNames]
		
		if any(header not in colsInFile for header in InReader.COL_HEADERS) :
			return False

		return True

	@staticmethod
	def verifyGenotype(genotype, seqLen, alphabetType, lineNum) :
		# Verfiy the length
		if len(genotype) != seqLen :
			print("Error: " + 
				ErrorCodes.getErrDescription(ErrorCodes.INCONSISTENT_SEQ_LEN)
				+ ": Line No. " + lineNum)

			raise GenonetsError(
					ErrorCodes.INCONSISTENT_SEQ_LEN,
					"Line No. " + lineNum)

		# Get the alphabet corresponding to the type received as
		# argument
		alphabet = SupportedAlphabet.getAlphabet(alphabetType)

		# Verify alphabet
		if any(letter not in alphabet for letter in genotype) :
			print("Error: " + 
				ErrorCodes.getErrDescription(ErrorCodes.ALPHABET_TYPE_MISMATCH)
				+ ": Line No. " + lineNum)

			raise GenonetsError(
					ErrorCodes.ALPHABET_TYPE_MISMATCH,
					"Line No. " + lineNum)


	# Open the given CSV file, and return a handle to the dict reader
	# for the file, as well as the file handle.
	@staticmethod
	def getDictReaderForFile(fileName) :
		# Open file
		try :
			dataFile = open(fileName, 'r')
		except Exception as e:
			print("Error: " + 
				ErrorCodes.getErrDescription(ErrorCodes.UNKNOWN_PARSING_ERROR))
			
			raise GenonetsError(ErrorCodes.UNKNOWN_PARSING_ERROR)

		# Read the file into a dictionary
		reader = csv.DictReader(dataFile, delimiter="\t")

		return reader, dataFile



#!/usr/bin/env python

"""
    genonets_exmpl_custom
    ~~~~~~~~~~~~~~~~~~~~~

    Illustrates the process of customizing the output by adding information to the
    result file that would not be added by Genonets by default.

    The 'Peaks' analysis is used as an example. By default, the 'Peaks' analysis 
    stores results in a dictionary of the form:
    {key=peakID : value=[genotypes in the peak]}. The sample code in this file
    customzes this dictionary by adding the score value corresponding to each
    genotype in the list. The resulting dictionary is of the format:
    {key=peakID : value=[(genotype1, score1), ..., (genotypeN, scoreN)]},
    i.e., it is a list of tuples.

	Use the following command to run the script:
	'python genonets_exmpl_custom.py DNA true data/genonets_sample_input.txt 0.35 results_custom'

	Output files will be generated in 'results_custom/'

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

from genonets import cmdl_handler					# For parsing command line arguments
from genonets import genonets_constants				# For analysis type constants
from genonets import genonets_interface as gn_if	# Interface to get the Genonets object

def process(args) :
	# Get a reference to the analysis constants. These constants are used
	# to specify analysis types.
	ac = genonets_constants.AnalysisConstants

	# Create the Genonets object. This will load the input file into
	# memory.
	gn = gn_if.Genonets(args)

	# Use 'gn' to create genotype networks for all genotype sets.
	gn.create()

	# Perform 'Peaks' analysis
	gn.analyze(analyses=[ac.PEAKS])

	# At this point, the analysis is done. We now need to extract the
	# data we need, i.e., the peaks dictionary.

	# For each genotype set,
	for genotypeSet in gn.getRepertoires() :
		# Get the igraph object for the giant
		giant = gn.getDominantNetFor(genotypeSet)

		# Get the dict of peaks {key=peakId : value=[list of sequences in the peak]}
		peaks = giant["Peaks"]

		# Update the dict of peaks, so that instead of just a list of
		# genotypes, we have a list of tuples (sequence, escore).
		newPeaks = {}

		# For each peak,
		for peak in peaks :
			# Initialize the list of tuples
			seqScrTuples = []

			# For each sequence in this peak,
			for sequence in peaks[peak] :
				# Find the corresponding vertex in the giant
				try :
					vertex = giant.vs.find(sequences=sequence)
				except ValueError :
					print("Oops! can't find " + sequence + " in giant.")

				# Get the escore
				score = giant.vs[vertex.index]["escores"]

				# Add the tuple to the list of tuples
				seqScrTuples.append((sequence, score))

			# Add the peak and the corresponding list of tuples to the
			# new peaks dict
			newPeaks[peak] = seqScrTuples

		# Replace the peaks dict in giant with the new dict. This is
		# useful because now if you use the genonets functions below to
		# save the network and results, the updated peaks dict with tuples
		# will be written automatically to file.
		giant["Peaks"] = newPeaks

	# Save networks to file in GML format
	gn.save()

	# Save the results to file from network level analysis
	gn.saveNetResults()


if __name__ == "__main__" :
	# Parse the command line arguments using the Genonets command line handler, and
	# pass the list of arguments to 'process()'.
	process(cmdl_handler.CmdParser().getArgs())

	# Print message to indicate processing is done.
	print("Done.\n")




"""
    robustness_functions
    ~~~~~~~~~~~~~~~~~~~~

    Contains functions used for genotype and phenotype robustness computations.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import numpy as np


class RobustnessAnalyzer :
	# Constructor
	def __init__(self, network, netBuilder) :
		# Reference to the network on which to perform this
		# analysis
		self.network = network

		# Get a reference to the NetworkUtils object
		self.netBuilder = netBuilder

		# Get a reference to the BitSeqManipulator in use
		self.bitManip = netBuilder.bitManip

		# Refernce to the list of tuples, (seq, robustness)
		# poupulated when required
		self.robustnessVals = None

	# Returns the arithmetic mean of the all values in the
	# robustness list.
	def getAvgRobustness(self) :
		# Calculate the mean of all robustness values for the network
		return np.mean(self.getRobustnessAll())

	# Returnes a list of tuples of the form (sequence, robustness) that
	# consist of robustness values for all sequences represented by
	# vertices of the given network.
	def getRobustnessAll(self, recompute=False) :
		# If either the robustness values have not been computed already,
		# or the caller has explicitly requested re-computing,
		if not self.robustnessVals or recompute :
			# Get robustness values for all sequences in the network
			self.robustnessVals = [self.getGenotypeRobustness(seq) \
									for seq in self.network.vs["sequences"]]

		return self.robustnessVals

	# Calculates sequence robustness, which is defined as Genotype
	# robustness in,
	# Ref: "Robustness and evolvability: a paradox resolved". Andreas 
	# Wagner, Proc. R. Soc. B 2008 275 91-100; 
	# DOI: 10.1098/rspb.2007.1137. Published 7 January 2008
	def getGenotypeRobustness(self, sequence) :
		# Get the vertex that corresponds to this sequence
		vertex = self.netBuilder.getVertex(sequence, self.network)
		
		# Get vertex degree
		degree = self.network.degree(vertex)

		# Get a list of all possible 1-neighbors for this sequence
		allNeighbors = self.bitManip.generateNeighbors(
			self.bitManip.seqToBits(sequence))

		# Count the number of all possible 1-neighbors of this sequence
		numNeighbors = len(allNeighbors)

		try :
			return float(degree) / float(numNeighbors)
		except ZeroDivisionError :
			return 0



#!/usr/bin/env python

"""
    genonets_exmpl_selective
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Demonstrates the steps required to create genotype networks, perform analyses, and
	write results to files, for a selected subset of genotype sets from the input file
	using Genonets.

	Use the following command to run the script:
	'python genonets_exmpl_selective.py DNA true data/genonets_sample_input.txt 0.35 results_selective'

	Output files will be generated in 'results_selective/'

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

from genonets import cmdl_handler					# For parsing command line arguments
from genonets import genonets_constants				# For analysis type constants
from genonets import genonets_interface as gn_if	# Interface to get the Genonets object

def process(args) :
	# Get a reference to the analysis constants. These constants are used
	# to specify analysis types.
	ac = genonets_constants.AnalysisConstants

	# Create the Genonets object. This will load the input file into
	# memory.
	gn = gn_if.Genonets(args)

	# Use 'gn' to create genotype networks for all genotype sets.
	gn.create()

	# Perform only 'Robustness' and 'Evolvability' analyses on just two of
	# the genotype sets available in the input file, i.e., 'Foxa2' and 'Bbx'.
	gn.analyze(["Foxa2", "Bbx"], analyses=[ac.ROBUSTNESS, ac.EVOLVABILITY])

	# Write the given genotype networks to files in GML format.
	# For a genotype network with two or more components, two files are generated: 
	# One corresponds to the entire network with all components, and the other 
	# corresponds to the dominant component only.
	gn.save(["Foxa2", "Bbx"])

	# Save genotype network level measures for the given genotype sets to 
	# 'Genotype_set_measures.txt'.
	gn.saveNetResults(["Foxa2", "Bbx"])

	# Save all genotype level measures for the given genotype sets to 
	# 'Foxa2_genotype_measures.txt' and 'Bbx_genotype_measures.txt' files.
	gn.saveGenotypeResults(["Foxa2", "Bbx"])


if __name__ == "__main__" :
	# Parse the command line arguments using the Genonets command line handler, and
	# pass the list of arguments to 'process()'.
	process(cmdl_handler.CmdParser().getArgs())

	# Print message to indicate processing is done.
	print("Done.\n")




"""
    accessibility_functions
    ~~~~~~~~~~~~~~~~~~~~~~~

    Exposes functions for accessibility, neighbor abundance, and diversity index computation.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

class AccessibilityAnalyzer :
	# Constructor
	def __init__(self, repertoire, network, repToGiantDict, dataDict, netBuilder, bitManip) :
		# Repertoire name
		self.repertoire = repertoire
		
		# Reference to the network on which to perform this
		# analysis
		self.network = network

		# Reference to dict: key=repertoire, value=network
		# FIXME: perhaps this dict should be replaced by the genonet
		#		 method getGraphFor() ...
		self.repToGiantDict = repToGiantDict

		# Reference to dict: key=rep, value=dict{seq: score}
		self.dataDict = dataDict

		# Get a reference to the NetworkBuilder object
		self.netBuilder = netBuilder

		# Reference to the bit manipulator object
		self.bitManip = bitManip

	# Computes the accessibility value for the given repertoire.
	# Ref: Cowperthwaite et al. (2008) PLoS Comp. Biol.
	def getAccessibility(self) :
		# Value to be calculated and returned
		accessibility = 0

		# Get the list of targets
		targets = self.dataDict.keys()

		# Remove the repertoire under consideration from the list of targets
		targets.remove(self.repertoire)

		# Get all sequences (in bit format) for the given repertoire
		sequences = [self.bitManip.seqToBits(seq) for seq in \
						self.network.vs["sequences"]]

		# For each target, compute F(j,i)
		for target in targets :
			# Get the genotype network for this target
			targetNet = self.repToGiantDict[target]

			# Get all external neighbors for this target
			extNeighbors = self.netBuilder.getAllExtNeighbors(targetNet)

			# Find sequences that exist in both extNeighbors and
			# sequences, i.e., intersection of both sets
			commonSeqs = set(extNeighbors) & set(sequences)

			# Calculate what fraction of sequences in extNeighbors
			# is in commonSeqs
			try :
				fraction = float(len(commonSeqs)) / float(len(extNeighbors))
			except ZeroDivisionError :
				fraction = 0

			# Add the fraction to the total accessibility value
			accessibility += fraction

		return accessibility

	# Computes the neighbor abundance for the given repertoire.
	# Ref: Cowperthwaite et al. (2008) PLoS Comp. Biol.
	def getNeighborAbundance(self) :
		# Value to be calculated and returned
		abundance = 0

		# Get all external neighbors for the given network
		extNeighbors = self.netBuilder.getAllExtNeighbors(self.network)

		# Get the list of targets
		targets = self.dataDict.keys()

		# Remove the repertoire under consideration from the list of targets
		targets.remove(self.repertoire)
		
		# For each target, compute F(i, j)
		for target in targets :
			# Get target sequences in bit format
			targetSeqs = [self.bitManip.seqToBits(seq) for seq in \
							self.repToGiantDict[target].vs["sequences"]]

			# Find sequences that exist in both extNeighbors and
			# targetSeqs, i.e., intersection of both sets
			commonSeqs = set(extNeighbors) & set(targetSeqs)

			# Calculate what fraction of targetSeqs in extNeighbors
			# is in commonSeqs
			try :
				fraction = float(len(commonSeqs)) / float(len(extNeighbors))
			except ZeroDivisionError :
				fraction = 0

			# Muliply the fraction with the target's abundance, i.e., the size
			# of the target genotype network
			abundanceFrac = float(fraction) * float(len(targetSeqs))

			# Add the fraction to the total abundance value
			abundance += abundanceFrac

		return abundance

	# Computes the E-statistic value for the given repertoire.
	# Ref: Cowperthwaite et al. (2008) PLoS Comp. Biol.
	def getPhenotypicDivesity(self) :
		# Value to be calculated and returned
		diversity = 0

		# Get all external neighbors for the given network
		extNeighbors = self.netBuilder.getAllExtNeighbors(self.network)

		# Get the list of targets
		targets = self.dataDict.keys()

		# Remove the repertoire under consideration from the list of targets
		targets.remove(self.repertoire)

		# For each target, compute F(i, j)
		for target in targets :
			# Get target sequences in bit format
			targetSeqs = [self.bitManip.seqToBits(seq) for seq in \
							self.repToGiantDict[target].vs["sequences"]]

			# Find sequences that exist in both extNeighbors and
			# targetSeqs, i.e., intersection of both sets
			commonSeqs = set(extNeighbors) & set(targetSeqs)

			# Calculate what fraction of targetSeqs in extNeighbors
			# is in commonSeqs
			try :
				fraction = float(len(commonSeqs)) / float(len(extNeighbors))
			except ZeroDivisionError :
				fraction = 0

			# Add the fraction to the total diversity value
			diversity += (fraction * fraction)
		
		return diversity




"""
    cmdl_handler
    ~~~~~~~~~~~~

    Handles command line arguments.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import argparse

from genonets_writer import Writer
from seq_bit_impl import BitManipFactory


# Parses the command line arguments using python's 'argparse' module
class CmdParser :
	# Constructor
	def __init__(self) :
		# Initialize the parser object
		parser = argparse.ArgumentParser()

		# Add 'alphabetType' as an argument
		parser.add_argument("alphabetType",
			help="Each genotype in the input file must be a string of letters from the alphabet selected here.", 
			choices=BitManipFactory.getMoleculeTypes())
		
		# Add 'includeIndels' as an argument
		parser.add_argument("includeIndels", 
			help="Should be set to 'True' if mutations that shift the entire genotype sequence by one letter should also be considered. Only single point mutations are considered when this parameter is set to 'False'.",
			choices=["True", "true", "False", "false"])
		
		# Add 'inFilePath' as an argument
		parser.add_argument("inFilePath", 
			help="Complete path to the input file, relative to the current directory.")
		
		# Add 'tau' as an argument
		parser.add_argument("tau", 
			help="Minimum score threshold: all genotypes in the input file with 'score' values below 'Tau' will be ignored. Tau must be a number.", 
			type=float)
		
		# Add 'outPath' as an argument
		parser.add_argument("outPath", 
			help="Path to the directory where output files should be generated. The directory will be created if it does not already exist.")
		
		# Keep an object level copy of the arguments dict
		self.args = parser.parse_args()

	# Returns parsed arguments
	def getArgs(self) :
		return self.args


# Stores command line arguments
class CmdArgs :
	# Constructor. Accepts a list of arguments.
	def __init__(self, arguments) :
		# Mandatory arguments

		# Molecule type: RNA, DNA, Protein, etc.
		self.moleculeType = arguments.alphabetType

		# Flag to indicate whether shift mutations should
		# be considered
		if arguments.includeIndels.lower() == "true" :
			self.useIndels = True
		else :
			self.useIndels = False

		# Path to the input file
		self.inFilePath = arguments.inFilePath

		# Lower bound on fitness values to be used.
		self.tau = arguments.tau

		# Path to the output folder
		self.outPath = arguments.outPath

		# Make sure the path ends with "/", since this is needed
		# in the file writing routines
		if not self.outPath.endswith("/") :
			self.outPath += "/"

		# Create a dictionary of parameters
		paramsDict = { 	"alphabetType"  : self.moleculeType,
					  	"includeIndels" : str(self.useIndels),
					  	"inFilePath"	: self.inFilePath,
					  	"tau"			: str(self.tau),
					  	"outPath"		: self.outPath }

		# Print the parsed parameter values
		self.printInParams(paramsDict)

		# Write input parameters to file
		Writer.writeInParamsToFile(paramsDict, self.outPath)

	# Print parsed input parameters
	def printInParams(self, paramsDict) :
		print("\nParsed input parameter values:")
		print("------------------------------")

		# For each parameter,
		for param in paramsDict.keys() :
			# Print the 'parameter : value' pair
			print(param + ": " + paramsDict[param])
		
		print("------------------------------\n")




"""
    genonets_interface
    ~~~~~~~~~~~~~~~~~~

    Public interface to Genonets functions.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import copy
from multiprocessing import Process, Queue

from cmdl_handler import CmdArgs
from genonets_writer import Writer
from genonets_reader import InReader
from graph_utils import NetworkBuilder
from seq_bit_impl import BitManipFactory
from genonets_filters import WriterFilter
from analysis_handler import AnalysisHandler
from genonets_constants import ErrorCodes
from genonets_exceptions import GenonetsError
from genonets_constants import GenonetsConstants as gc
from genonets_constants import AnalysisConstants as ac


class Genonets :
	ALL = 0

	# Constructor
	def __init__(self, args) :
		# Handle program arguments
		self.cmdArgs = CmdArgs(args)

		# Read file and load input data into memory
		self.inDataDict, self.deltaDict, self.seqToRepDict, self.seqLength = \
				self.buildDataDicts(self.cmdArgs.inFilePath)

		# Get the bit-sequence manipulator object corresponding to the
		# given molecule type.
		self.bitManip = self.getBitManip()

		# Dict {bitseq : seq}
		# The function call uses bitManip, so bitManip must be created first
		self.bitsToSeqDict = self.buildBitsToSeqDict()

		# Get the NetworkUtils object
		self.netBuilder = NetworkBuilder(self.bitManip)

		# Dictionary: Key=Repertoire, Value=Network. Created when required.
		self.repToNetDict = {}

		# Dictionary: Key=Repertoire, Value=Giant. Created when required.
		# If there is only one component, giant=network
		self.repToGiantDict = {}

		# Create the analyzer object
		self.analyzer = AnalysisHandler(self)

	# ----------------------------------------------------------------------
	#	Public interface
	# ----------------------------------------------------------------------

	# Description: 	Creates genotype networks for the given list of genotype set names.
	# Arguments:
	#	'repertoires': 	List of genotype set names. This argument is optional, and
	#					if absent, results in the creation of all available 
	#					genotype networks.
	#	'parallel':		Flag to indicate whether or not parallel processing should
	#					be used. This argument is optional, and is set to 'False' 
	#					by default.
	# Return:		No return value.
	def create(self, repertoires=gc.ALL, parallel=False) :
		# If a single string is received, convert it into an iterable
		repertoires = [repertoires] if type(repertoires) == str else repertoires

		# If all repertoires should be considered,
		if repertoires == gc.ALL :
			# Get a list of all repertoires
			repertoires = self.getRepertoires()

		# If multiprocessing should be used,
		if parallel == True:			
			self.createNets_parallel(repertoires)
		else :
			self.createNets(repertoires)

	# Description:	Performs the given set of analyses on the given list of
	#				genotype sets.
	# Arguments:
	#	'repertoires': 	List of genotype set names. This argument is optional, and
	#					if absent, results in the processing of all available 
	#					genotype networks.
	#	'analyses':		List of analysis types. This argumetn is optional. If the
	#					the argument is not provided, all available analyses are
	#					performed.
	#	'parallel':		Flag to indicate whether or not parallel processing should
	#					be used. This argument is optional, and is set to 'False' 
	#					by default.
	# Return:		No return value.
	def analyze(self, repertoires=gc.ALL, analyses=gc.ALL, parallel=False) :
		# If all repertoires should be considered,
		if repertoires == gc.ALL :
			# Get a list of all repertoires
			repertoires = self.getRepertoires()

		# If a single string is received, convert it into an iterable
		repertoires = [repertoires] if type(repertoires) == str else repertoires

		# If overlap in one of the requested analyses, there need to be at
		# at least two repertoires in the dataset
		if analyses == gc.ALL or ac.OVERLAP in analyses :
			if len(repertoires) < 2 :
				print("Error: " + 
					ErrorCodes.getErrDescription(ErrorCodes.NOT_ENOUGH_REPS_OLAP)
					+ ": Tau=" + str(self.cmdArgs.tau))

				raise GenonetsError(
						ErrorCodes.NOT_ENOUGH_REPS_OLAP,
						"Tau=" + str(self.cmdArgs.tau))

		# If multiprocessing should be used,
		if parallel == True :
			# Perform all analyses in parallel; overlap will be ignored.
			self.analyzeNets_parallel(repertoires, analyses)

			if analyses == gc.ALL or ac.OVERLAP in analyses :
				# Reset analysis handler to make sure it references
				# the updated dicts
				del self.analyzer
				self.analyzer = AnalysisHandler(self)

				# Use serial processing to perform overlap analysis
				self.analyzeNets(repertoires, [ac.OVERLAP])

				print("Finished performing overlap analysis.")
		else :
			# Perform all analyses using serial processing
			self.analyzeNets(repertoires, analyses)

	# Description:	Creates the phenotype network from the given list of
	#				genotype sets.
	# Arguments:
	#	'collection':	Name to give the phenotype network.
	#	'repertoires': 	List of genotype set names. This argument is optional, and
	#					if absent, results in the processing of all available 
	#					genotype networks.
	# Return:		igraph object corresponding to the phenotype network.
	# TODO: Add a check to make sure the evolvability analysis has been done
	#		for all networks that are to be processed here ...
	def getEvoNet(self, collection="Species", repertoires=gc.ALL) :
		# If a single string is received, convert it into an iterable
		repertoires = [repertoires] if type(repertoires) == str else repertoires

		# If all repertoires should be considered,
		if repertoires == gc.ALL :
			# Get a list of all repertoires
			repertoires = self.getRepertoires()

		# Create a list of giants for the given repertoires
		giants = [self.repToGiantDict[repertoire] for repertoire in repertoires]

		# Create the evolvability network, and get the igraph object
		evoNet = self.netBuilder.createEvoNet(collection, giants)

		return evoNet

	# Description:	Returns the list of names of all genotype sets for which
	#				genotype networks have been created.
	# Return:		List of names of genotype sets.
	def getRepertoires(self) :
		repertoires = self.inDataDict.keys()
		repertoires = [repertoires] if type(repertoires) == str else repertoires

		return repertoires

	# Description:	Returns the igraph object for the network corresponding to the
	#				given genotype set name.
	# Arguments:
	#	'repertoire':	Name of the genotype network for which the igraph object
	#					is required.
	# Return:		igraph object for the required network.
	def getNetworkFor(self, repertoire) :
		try :
			return self.repToNetDict[repertoire]
		except KeyError :
			return None

	# Description:	Returns the igraph object for the giant component 
	#				corresponding to the given genotype set name.
	# Arguments:
	#	'repertoire':	Name of the genotype network for which the igraph object
	#					is required.
	# Return:		igraph object for the giant component corresponding to the
	#				given genotype set name.
	def getDominantNetFor(self, repertoire) :
		try :
			return self.repToGiantDict[repertoire]
		except KeyError :
			return None

	# Description:	Returns the overlap matrix for all the genotype networks.
	# Return:		Overlap matrix as a list of lists.
	def getOverlapMat(self) :
		# Overlap matrix can only be computed if the networks have already
		# been created.
		if len(self.repToGiantDict) == 0 :
			# Networks have not been created. Therefore, overlap matrix
			# cannot be computed.
			print("Overlap matrix cannot be computed before network creation.")

			return None

		# If the overlap matrix has already been computed,
		if self.analyzer.overlapMatrix :
			# No need to compute again, just return the existing matrix
			return self.analyzer.overlapMatrix
		else :
			# Perform the overlap compution
			self.analyzer.overlap()

			# Return the resulting matrix
			return self.analyzer.overlapMatrix

	# Description:	Save the given genotype networks as GML files.
	# Arguments:
	#	'repertoires':	List of names of genotype networks to be saved to
	#					file.
	def save(self, repertoires=gc.ALL) :
		# If a single string is received, convert it into an iterable
		repertoires = [repertoires] if type(repertoires) == str else repertoires
		
		Writer.writeNetsToFile(self.repToNetDict, self.repToGiantDict, 
			self.netBuilder, self.cmdArgs.outPath, 
			WriterFilter.gmlAttribsToIgnore, repertoires)

	# Description:	Writes the network level results to file.
	# Arguments:
	#	'repertoires':	List of names of genotype networks for which results
	#					need to be written to file.
	def saveNetResults(self, repertoires=gc.ALL) :
		# If a single string is received, convert it into an iterable
		repertoires = [repertoires] if type(repertoires) == str else repertoires
		
		Writer.writeNetAttribs(self.repToNetDict, self.repToGiantDict, 
			self.netBuilder, self.cmdArgs.outPath,
			WriterFilter.netAttribsToIgnore,  repertoires)

	# Description:	Writes the genotype level results to file.
	# Arguments:
	#	'repertoires':	List of names of genotype networks for which results
	#					need to be written to file.
	def saveGenotypeResults(self, repertoires=gc.ALL) :
		# If a single string is received, convert it into an iterable
		repertoires = [repertoires] if type(repertoires) == str else repertoires

		Writer.writeSeqAttribs(self.repToNetDict, self.repToGiantDict, 
			self.netBuilder, self.cmdArgs.outPath,
			WriterFilter.seqAttribsToIgnore, repertoires)

	# Description:	Save the phenotype network as a GML file.
	# Arguments:
	#	'phenotypeNetwork':	igraph object corresponding to the phenotype 
	#						network.
	def saveEvoNet(self, phenotypeNetwork) :
		Writer.writeNetToFile(phenotypeNetwork, self.cmdArgs.outPath, 
			WriterFilter.gmlAttribsToIgnore)

	# Plots the given network.
	def plot(self, network, layout="auto") :
		self.netBuilder.plotNetwork(network, layout, self.cmdArgs.outPath)

	# ----------------------------------------------------------------------
	#	Private methods, i.e., those that are not supposed to be part of
	#	public interface
	# ----------------------------------------------------------------------

	def buildDataDicts(self, inFilePath) :
		return InReader.buildDataDicts(inFilePath, self.cmdArgs.tau, 
				self.cmdArgs.moleculeType)

	def getBitManip(self) :
		return BitManipFactory.getBitSeqManip(self.cmdArgs.moleculeType, 
				self.seqLength, self.cmdArgs.useIndels)

	def getBitSeqsAndScores(self, repertoire) :
		# Get the list of sequences for the given repertoire
		sequences = self.inDataDict[repertoire].keys()

		# Get the list of scores for the given repertoire
		scores = [self.inDataDict[repertoire][seq] for seq in sequences]

		return sequences, scores

	def buildBitsToSeqDict(self) :
		# Get the list of available repertoires
		repertoires = self.getRepertoires()

		# Construct a list of all sequences available in the input data
		allSeqs = []

		# For each repertoire,
		for repertoire in repertoires :
			# Get the list of all sequences in the repertoire
			allSeqs.extend(self.inDataDict[repertoire].keys())

		# Remove all redundant entries in the list
		uniqueSeqs = list(set(allSeqs))

		# Construct the dict {bitseq : seq}
		return { self.bitManip.seqToBits(seq) : seq \
					for seq in uniqueSeqs }

	# Create genotype networks for the given, or all repertoires
	def createNets(self, repertoires) :
		# For each repertoire,
		for repertoire in repertoires :
			# Get the sequences and scores
			seqs, scores = self.getBitSeqsAndScores(repertoire)

			# Create the genotype network and store it in a 
			# dictionary: Key=Repertoire, Value=Network
			self.repToNetDict[repertoire] = \
				self.netBuilder.createGenoNet(repertoire, seqs, scores)

			# Get the number of components in the network
			numComponents = len(self.netBuilder.getComponents( \
				self.repToNetDict[repertoire]))

			# If there are more than one components,
			if numComponents > 1 :
				# Get the giant component
				giant = self.netBuilder.getGiantComponent(self.repToNetDict[repertoire])

				# Set 'name' attribute for giant
				giant["name"] = repertoire + "_dominant"

				# Reference to the giant component
				self.repToGiantDict[repertoire] = giant
			else :
				# The network and giant component are the same
				self.repToGiantDict[repertoire] = self.repToNetDict[repertoire]

	# Use multiprocessing to create genotype networks
	# for the given, or all repertoires
	def createNets_parallel(self, repertoires) :
		# Instantiate a concurrent queue for results
		resultsQueue = Queue()

		# Create separate processes for each repertoire
		processes = [ Process(  target=Genonets.createGN, \
								args=( self.inDataDict[repertoire], \
									   self.cmdArgs, \
									   self.seqLength, \
									   resultsQueue, \
									   repertoire ) \
							 ) \
						for repertoire in repertoires ]

		# Start the processses
		for p in processes :
			p.start()

		# Spin lock
		# FIXME: The condtion in the loop can result in an infinite
		#		 iteration if one of the processes does not put results
		#		 in the queue. This condition should be replaced
		#		 with one that is reliable ...
		while len(self.repToNetDict) != len(repertoires) :
			result = resultsQueue.get()

			self.repToNetDict[result[0]] = result[1][0]
			self.repToGiantDict[result[0]] = result[1][1]

	@staticmethod
	def createGN(seqScrDict, args, seqLength, resultsQueue, repertoire) :
		# Get a reference to the bit manipulator
		bitManip = BitManipFactory.getBitSeqManip(args.moleculeType, 
					seqLength, args.useIndels)

		# Get the sequences for the given repertoire
		sequences = seqScrDict.keys()

		# Get the list of scores for the given repertoire
		scores = [seqScrDict[sequence] for sequence in sequences]

		# Create a network builder object
		netBuilder = NetworkBuilder(bitManip)

		# Create the genotype network
		network = netBuilder.createGenoNet(repertoire, sequences, scores)

		# Get the number of components in the network
		numComponents = len(netBuilder.getComponents(network))

		# If there are more than one components,
		if numComponents > 1 :
			# Get the giant component
			giant = netBuilder.getGiantComponent(network)

			# Set 'name' attribute for giant
			giant["name"] = repertoire + "_dominant"
		else :
			# The network and giant component are the same
			giant = network

		# Create a tuple in which to put the results
		netTuple = (repertoire, (network, giant))

		# Add the created network objects to the shared queue		
		resultsQueue.put(netTuple)

		# Close the queue for this process
		resultsQueue.close()

	def analyzeNets(self, repertoires, analyses) :
		# For each repertoire,
		for repertoire in repertoires :
			# Perform the analysis
			self.analyzer.analyze(repertoire, analyses)

	def analyzeNets_parallel(self, repertoires, analyses) :
		# Instantiate a concurrent queue for results
		resultsQueue = Queue()

		# Create separate processes for each repertoire
		processes = [ Process(  target=Genonets.analyzeGN, \
								args=( copy.deepcopy(self), \
									   analyses, \
									   resultsQueue, \
									   repertoire ) \
							 ) \
						for repertoire in repertoires ]

		# Start the processses
		for p in processes :
			p.start()

		# Delete the existing net dicts
		del self.repToNetDict
		del self.repToGiantDict

		# Re-initialize the deleted dicts
		self.repToNetDict = {}
		self.repToGiantDict = {}

		# Spin lock
		# FIXME: The condtion in the loop can result in an infinite
		#		 iteration if one of the processes does not put results
		#		 in the queue. This condition should be replaced
		#		 with one that is reliable ...
		while len(self.repToNetDict) != len(repertoires) :
			result = resultsQueue.get()

			print("Analysis results received for: " + result[0])

			self.repToNetDict[result[0]] = result[1][0]
			self.repToGiantDict[result[0]] = result[1][1]

	@staticmethod
	def analyzeGN(genonetsCopy, analyses, resultsQueue, repertoire) :
		# Initialize the AnalysisHandler object
		analyzer = AnalysisHandler(genonetsCopy, parallel=True)

		# Perform the analyses
		analyzer.analyze(repertoire, analyses)

		# Create a tuple in which to put the results
		resultTuple = ( repertoire, (genonetsCopy.getNetworkFor(repertoire), \
					 genonetsCopy.getDominantNetFor(repertoire)) )

		# Add results to the shared queue
		resultsQueue.put(resultTuple)

		# Close the queue for this process
		resultsQueue.close()




"""
    landscape_functions
    ~~~~~~~~~~~~~~~~~~~

    Wrapper for all landscape analysis functions.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import igraph

from peak_functions import PeakAnalyzer
from path_functions import PathAnalyzer
from epistasis_functions import EpistasisAnalyzer
from genonets_utils import Utils


class Landscape :
	# Constructor
	def __init__(self, network, netUtils, seqToEscrDict, delta, bitManip) :
		# Store reference to the network object
		self.network = network

		# Store reference to the netUtils object
		self.netUtils = netUtils
		
		# Create the peak analyzer
		self.peakAnalyzer = PeakAnalyzer(network, netUtils, delta)

		# Create the path analyzer
		self.pathAnalyzer = PathAnalyzer(network, netUtils, delta)

		# Create the epistasis analyzer
		self.epiAnalyzer = EpistasisAnalyzer(network, netUtils, 
			seqToEscrDict, delta, bitManip)

		# Get a reference to the BitSeqManipulator in use
		self.bitManip = self.netUtils.bitManip

	# ----------------------------------------------------------------
	# Peak analysis methods
	# ----------------------------------------------------------------
	
	def getPeaks(self, recompute) :
		return self.peakAnalyzer.getPeaks(recompute)

	# ----------------------------------------------------------------
	# Path analysis methods
	# ----------------------------------------------------------------

	def getAccessiblePaths(self, pathLength) :
		return self.pathAnalyzer.getAccessiblePaths(pathLength)

	# ----------------------------------------------------------------
	# Epistasis analysis methods
	# ----------------------------------------------------------------

	def getEpistasis(self) :
		return self.epiAnalyzer.getEpiAll()

	# ----------------------------------------------------------------
	# Calculate mutational distances from summit for all vertices
	# ----------------------------------------------------------------

	def populateDistsToSummit(self) :
		# Convenient handle for bitManip
		bm = self.bitManip

		# Get the summit sequence
		summit = Utils.getSeqWithMaxScore(self.network, 
			self.bitManip.seqLength)

		# Get vertex that represents summit
		trgtVrtx = self.netUtils.getVertex(summit, self.network)

		# Reference to the list of sequences in the network
		vertices = [ self.netUtils.getVertex(seq, self.network) \
					 for seq in self.network.vs["sequences"] ]

		self.network.vs["Distance from Summit"] = \
			[ 	len(self.network.get_shortest_paths(srcVrtx, to=trgtVrtx, \
						weights=None, mode=igraph.OUT, output="epath")[0]) \
			  	for srcVrtx in vertices ]




"""
    analysis_handler
    ~~~~~~~~~~~~~~~~

    Exposes wrapper functions, one per analysis type. Serves as a collective interface to handler classes for all analysis types.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import json	# For proper list stringification

from genonets_writer import Writer
from landscape_functions import Landscape
from overlap_functions import OverlapAnalyzer
from structure_functions import StructureAnalyzer
from robustness_functions import RobustnessAnalyzer
from evolvability_functions import EvolvabilityAnalyzer
from accessibility_functions import AccessibilityAnalyzer
from genonets_constants import AnalysisConstants as ac
from genonets_constants import EpistasisConstants as epi


class AnalysisHandler :
	# Constructor
	def __init__(self, caller, parallel=False) :
		# Store a reference to the caller object
		self.caller = caller

		# Set the flag that indicates whether or not this instance
		# is being used for parallel processing
		self.parallel = parallel

		# Dictionary: Key=Repertoire, Value=Network. Created when required.
		self.repToNetDict = self.caller.repToNetDict

		# Dictionary: Key=Repertoire, Value=Giant. Created when required.
		self.repToGiantDict = self.caller.repToGiantDict

		# Reference to the NetBuilder object in use
		self.netBuilder = self.caller.netBuilder

		# Get the bit-sequence manipulator object corresponding to the
		# given molecule type.
		self.bitManip = self.caller.bitManip

		# Reference to input data dictionary
		self.inDataDict = self.caller.inDataDict

		# Reference to delta dict
		self.deltaDict = self.caller.deltaDict

		# Reference to 
		self.seqToRepDict = self.caller.seqToRepDict

		# Dict {bitseq : seq}
		self.bitsToSeqDict = self.caller.bitsToSeqDict

		# Dictionary to store 'analysis type' to 'function'
		# mapping
		self.analysisToFunc = {	#ac.LANDSCAPE : self.landscape, \
								ac.PEAKS : self.peaks, \
								ac.PATHS : self.paths, \
								ac.EPISTASIS : self.epistasis,\
								ac.ROBUSTNESS : self.robustness, \
								ac.EVOLVABILITY : self.evolvability, \
								ac.ACCESSIBILITY : self.accessibility, \
								ac.NEIGHBOR_ABUNDANCE : self.neighborAbundance, \
								ac.PHENOTYPIC_DIVERSITY : self.phenotypicDiversity, \
								ac.STRUCTURE : self.structure, \
								ac.OVERLAP : self.overlap }

		# Refernce to the overlap matrix
		self.overlapMatrix = None

	def getFuncsFor(self, analysis) :
		try :
			funcs = self.analysisToFunc[analysis]

			return funcs if type(funcs) == list else [funcs]
		except KeyError :
			return None

	def analyze(self, repertoire, analyses=ac.ALL) :
		if analyses == ac.ALL :
			analyses = self.analysisToFunc.keys()

		# For each analysis type specified in the list,
		for analysis in analyses :
			# Get a list of function names
			functions = self.getFuncsFor(analysis)

			# For each function name,
			for function in functions :
				# Call the function for the given repertoire
				function(repertoire)

	def getLandscapeObj(self, giant, repertoire) :
		# Get the landscape object
		lscape = Landscape(giant, self.netBuilder, self.inDataDict[repertoire], 
			self.deltaDict[repertoire], self.bitManip)

		return lscape

	def peaks(self, repertoire) :
		# Get the dominant genotype network for the repertoire
		giant = self.caller.getDominantNetFor(repertoire)

		# Get the landscape object
		lscape = self.getLandscapeObj(giant, repertoire)

		# Get peaks
		peaks = lscape.getPeaks(recompute=True)

		# Set the computed values as a network attribute
		giant["Number_of_peaks"] = len(peaks)

		# Store a dict - {key=peakId, value=[sequences in the peak]}
		giant["Peaks"] = { peakId : peaks[peakId]["sequences"] \
							for peakId in peaks.keys() }

		# For each vertex, add distance to summit as a vertex attribute
		lscape.populateDistsToSummit()

	def paths(self, repertoire) :
		# Get the dominant genotype network for the repertoire
		giant = self.caller.getDominantNetFor(repertoire)

		# Get the landscape object
		lscape = self.getLandscapeObj(giant, repertoire)

		# Get paths
		lscape.getAccessiblePaths(0)

		# Set the computed value as a network attribute
		giant["Summit"] = lscape.pathAnalyzer.getSummitId()

		# Vertex level attributes
		allPathsToPeak = lscape.pathAnalyzer.getAllPathsToPeak()
		giant.vs["pathsToSummit"] = [ allPathsToPeak[i] \
										for i in range(len(allPathsToPeak)) ]

		# Add the count for each vertex as a vertex level attribute in giant
		giant.vs["Accessible_paths_through"] = lscape.pathAnalyzer.getPathsThruVtxs()

	def epistasis(self, repertoire) :
		# Get the dominant genotype network for the repertoire
		giant = self.caller.getDominantNetFor(repertoire)

		# Get the landscape object
		lscape = self.getLandscapeObj(giant, repertoire)

		# Get epistasis
		epistasis = lscape.getEpistasis()

		# Get the vertex to squares dict
		vtxToSqrs, squares = lscape.epiAnalyzer.getVertexToSquaresDict()

		# Set the computed values as network attributes
		giant["Squares_list"] = json.dumps(squares)
		giant["SqrEpi_list"] = json.dumps(lscape.epiAnalyzer.getSqrEpi())
		giant["Number_of_squares"] = len(lscape.epiAnalyzer.squares)
		giant["Magnitude_epistasis"] = epistasis[epi.MAGNITUDE]
		giant["Simple_sign_epistasis"] = epistasis[epi.SIGN]
		giant["Reciprocal_sign_epistasis"] = epistasis[epi.RECIPROCAL_SIGN]

		# Vertex level attributes

		# Can't traverse an empty dict
		if len(vtxToSqrs) > 0 :
			# Since vtxToSqrs is a dict, it is important to convert it into an ordered
			# list for correct mapping to vertices
			giant.vs["VtxToSqrs"] = [ vtxToSqrs[i] for i in range(len(vtxToSqrs)) ]
		else :
			# Empty list. This makes it possible for the user to distinguish between
			# no squares associated with a vertex, and epistasis function was not
			# called at all
			# Note: Has to be a nested empty list, since otherwise igraph throws an
			# exception
			giant.vs["VtxToSqrs"] = [[]]

	def robustness(self, repertoire) :
		# Get the dominant genotype network for the repertoire
		giant = self.caller.getDominantNetFor(repertoire)

		# Construct a RobustnessAnalyzer object
		robAnalyzer = RobustnessAnalyzer(giant, self.netBuilder)

		# Compute repertoire robustness and set it as a network
		# attribute
		giant["Robustness"] = robAnalyzer.getAvgRobustness()

		# Set robustness values for all vertices, i.e., sequences
		giant.vs["Robustness"] = robAnalyzer.getRobustnessAll()

	def evolvability(self, repertoire) :
		# Get the dominant genotype network for the repertoire
		giant = self.caller.getDominantNetFor(repertoire)

		# Construct a EvolvabilityAnalyzer object
		evoAnalyzer = EvolvabilityAnalyzer(giant, self.inDataDict, 
			self.seqToRepDict, self.repToGiantDict, self.netBuilder,
			self.bitsToSeqDict)

		# Compute repertoire evolvability and set it as a network
		# attribute
		repertoireEvo, targetRepertoires = evoAnalyzer.getReportoireEvo()
		giant["Evolvability"] = repertoireEvo
		
		# Stringify the list, since pythons lists cannot be written to GML.
		giant["Evolvability_targets"] = json.dumps(targetRepertoires)

		# Set evolvability values for all vertices, i.e., sequences
		evoTuples = evoAnalyzer.getEvoAll()
		
		evoScores  = [evoTuples[i][0] for i in range(len(evoTuples))]
		evoTargets = [evoTuples[i][1] for i in range(len(evoTuples))]

		giant.vs["Evolvability"] = evoScores
		giant.vs["Evolves_to_genotypes_in"] = [evoTargets[i].keys() \
												for i in range(len(evoTargets))]
		giant.vs["Evolvability_targets"] = evoTargets

	def accessibility(self, repertoire) :
		# Get the dominant genotype network for the repertoire
		giant = self.caller.getDominantNetFor(repertoire)

		# Create an AccessibilityAnalyzer object
		accAnalyzer = AccessibilityAnalyzer(repertoire, giant, self.repToGiantDict,
			self.inDataDict, self.netBuilder, self.bitManip)

		# Compute repertoire accessibility and set it as a network attribute
		giant["Accessibility"] = accAnalyzer.getAccessibility()

	def neighborAbundance(self, repertoire) :
		# Get the dominant genotype network for the repertoire
		giant = self.caller.getDominantNetFor(repertoire)

		# Create an AccessibilityAnalyzer object
		accAnalyzer = AccessibilityAnalyzer(repertoire, giant, self.repToGiantDict,
			self.inDataDict, self.netBuilder, self.bitManip)

		# Compute repertoire neighborhood abundance and set it as a network
		# attribute
		giant["Neighbor_abundance"] = accAnalyzer.getNeighborAbundance()

	def phenotypicDiversity(self, repertoire) :
		# Get the dominant genotype network for the repertoire
		giant = self.caller.getDominantNetFor(repertoire)

		# Create an AccessibilityAnalyzer object
		accAnalyzer = AccessibilityAnalyzer(repertoire, giant, self.repToGiantDict,
			self.inDataDict, self.netBuilder, self.bitManip)

		# Compute phenotypic diversity and set it as a network attribute
		giant["Phenotypic_diversity"] = accAnalyzer.getPhenotypicDivesity()

	def structure(self, repertoire) :
		# Get the genotype network for the repertoire
		network = self.caller.getNetworkFor(repertoire)

		# Get the dominant genotype network for the repertoire
		giant = self.caller.getDominantNetFor(repertoire)

		# Create the structure analyzer object
		structAnalyzer = StructureAnalyzer(network, self.netBuilder)

		# Compute and set network/giant level properties
		network["Genotype_network_sizes"] = str(structAnalyzer.getComponentSizes())
		network["Number_of_genotype_networks"] = structAnalyzer.getNumComponents()
		network["Size_of_dominant_genotype_network"] = structAnalyzer.getDominantSize()
		network["Proportional_size_of_dominant_genotype_network"] = structAnalyzer.getPercentDominantSize()
		giant["Edge_density"] = structAnalyzer.getEdgeDensity()
		giant["Diameter"] = structAnalyzer.getDiameter()		
		giant["Average_clustering_coefficient_of_dominant_genotype_network"] = structAnalyzer.getAvgClstrCoeff()
		giant["Assortativity"] = structAnalyzer.getAssortativity()
		
		# The list of vertex Ids needs to be stringified, since otherwise these
		# cannot be written to GML.
		giant["diameterPath_list"] = json.dumps(structAnalyzer.getDiameterPath())

		# Compute and set vertex level properties
		giant.vs["Coreness"] = structAnalyzer.getCoreness()
		giant.vs["Clustering_coefficient"] = structAnalyzer.getClusteringCoefficients()

	def overlap(self, r=None) :
		# Overlap analysis is not allowed during parallel processing,
		# since it needs to be performed only once
		if self.parallel == True :
			return

		# If overlap has been computed for any of the repertoires already,
		# it does not need to be computed again.
		if not self.overlapMatrix :
			# Creat the overlap analyzer
			overlapAnalyzer = OverlapAnalyzer(self.repToGiantDict,
				self.caller.getRepertoires())

			# Compute overlap data
			self.overlapMatrix, repertoires, overlapDict = overlapAnalyzer.getOverlapData()

			# If the overlap dict was populated
			if overlapDict :
				# Use the overlap dict to populate vertex level attributes in all giants

				# For each repertoire,
				for repertoire in self.caller.getRepertoires() :
					# Get giant
					giant = self.repToGiantDict[repertoire]

					# Get the sequence dict for this repertoire
					seqDict = overlapDict[repertoire]

					# For each sequence in seq dict,
					for sequence in seqDict.keys() :
						# Get the corresponding vertex id from the network
						vertex = self.netBuilder.getVertex(sequence, giant)

						# Add the list of targets as an attribute for this
						# vertex in giant
						giant.vs[vertex.index]["Overlaps_with_genotypes_in"] = \
									seqDict[sequence]

			# If overalp matrix was populated,
			if self.overlapMatrix :
				# Write matrix to file
				Writer.writeOverlapToFile(self.overlapMatrix, repertoires, 
						self.caller.cmdArgs.outPath)




"""
    graph_utils
    ~~~~~~~~~~~

    Function for network manipulation.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import math
import json
import igraph


# Provides methods for constructing and manipulating genotype
# networks. The methods in this class are agnostic to molecule
# type.
class NetworkBuilder :
	# Constructor
	def __init__(self, seqBitManip) :
		# Store reference to the sequence/bit manipulator object
		# in use
		self.bitManip = seqBitManip

	# Return 'True' if the source sequence is connected to the target
	# sequence, and 'False' otherwise.
	def areConnected(self, source, target) :
		return self.bitManip.areNeighbors( \
					self.bitManip.seqToBits(source), \
					self.bitManip.seqToBits(target))

	# Create a genotype network using the give list of sequences
	def createGenoNet(self, repertoire, sequences, scores) :
		# Initialize the empty network
		network = igraph.Graph()

		# Set repertoire name as a graph attribute
		network["name"] = repertoire

		# Create vertices
		network.add_vertices(len(sequences))

		# Add scores to all vertices
		# FIXME: Change back from 'escores' to 'scores'
		network.vs["escores"]	= scores
		
		# Add sequences to all vertices
		network.vs["sequences"] = sequences
		
		# Add bit formatted sequences to all vertices
		# Note: For very long sequences, the size of the corresponding
		#		integer value becomes too long for igraph to be able
		#		to deal with it (since igraph is C based). Therefore,
		#		bitseqs cannot be stored in the graph.
		bitseqs = [self.bitManip.seqToBits(seq) for seq in sequences]
	
		# Get a list of pairs of indices which represent paris of
		# sequences that are 1-neighbors, i.e., only one mutation
		# apart.
		edges = [(i,j) \
					for i in range(len(bitseqs) - 1) \
					for j in range(i + 1, len(bitseqs)) \
					if self.bitManip.areNeighbors(bitseqs[i], bitseqs[j])]

		# Connect the two with an edge
		network.add_edges(edges)
		
		return network

	# For the given sequence, returns a list of 1-neighbors that are not part of
	# the given genotype network.
	# Note: The list of external neighbors returned comprises bitseqs, not
	#		strings. The caller must be aware of this.
	def getExternalNeighbors(self, sequence, network) :
		# Get a list of all possible 1-neighbors of the given sequence
		allNeighbors = [ neighbor for neighbor in \
			self.bitManip.generateNeighbors(self.bitManip.seqToBits(sequence)) ]

		# Get a list of 1-neighbors that are part of the given genotype
		# network
		genoNetNeighbors = [self.bitManip.seqToBits(network.vs[neighbor]["sequences"]) \
								for neighbor in self.getNeighbors(sequence, network)]

		# Get a list of sequences that are 1-neighbors, but not part of the
		# genotype network. This is achieved by taking the difference of
		# two sets, i.e., elements in allNeighbors not in genoNetNeighbors
		externNeighbors = list(set(allNeighbors) - set(genoNetNeighbors))

		return externNeighbors

	# Get all unique external neighbors for the given genotype network
	def getAllExtNeighbors(self, network) :
		# List to be populated with external neighbors
		extNeighbors = []

		# Get all sequences
		sequences = network.vs["sequences"]

		# For each sequence, get the list of external neighbors and 
		# append it to extNeighbors
		for sequence in sequences :
			extNeighbors.extend(self.getExternalNeighbors(sequence, network))

		# Remove redundant sequences
		extNeighbors = list(set(extNeighbors))

		return extNeighbors

	# Get the vertex based on the sequence string from the
	# given network
	def getVertex(self, sequence, network) :
		try :
			return network.vs.find(sequences=sequence)
		except ValueError :
			print("Error! ... Non-existent vertex requested: " + str(sequence))

	# Returns a list of sequences that constitute the neighborhood
	# of the given sequence within the given network.
	def getNeighbors(self, sequence, network) :
		return network.neighbors(self.getVertex(sequence, network))

	# Plot the given network using the given layout.
	def plotNetwork(self, network, layout, outPath) :
		vertices = [network.vs[i].index for i in range(0, network.vcount())]
		# For GNs,
		# degrees = network.degree(vertices)
		# For ENs,
		degrees = network.outdegree(vertices)

		visual_style = {}
		# For GNs:
		#visual_style["vertex_size"] = [x*2 for x in degrees]
		# For ENs:
		#visual_style["vertex_size"] = [x*2 for x in degrees]
		visual_style["vertex_size"] = [(x*10 * x*10) for x in network.vs["Evolvability"]]
		#visual_style["edge_width"] = [math.log(x) for x in network.es["weight"]]
		#visual_style["vertex_color"] = [int(x*300) for x in network.vs["escores"]]

		layout = network.layout(layout)

		igraph.plot(network, outPath + network["name"] + "_evonet.svg", bbox=(1500, 1500), **visual_style)

	# Return a list, the size of which represents the number of commnected
	# component in the network. Each element in the list corresponds to the
	# size of one of the connected components.
	def getComponents(self, network) :
		vertexCluster = network.components()

		return vertexCluster.sizes()

	# Returns the gian component.
	def getGiantComponent(self, network) :
		return network.components().giant()

	# Create a network that consists of the given sequence and its
	# one-neighbors.
	def createNeighborhoodNet(self, sequence, network) :
		sequences = [sequence]
		escores = []

		# Get the vertex corresponding to the sequence
		vertex = self.getVertex(sequence, network)

		# Get all adjacent vertices
		neighbors = network.neighbors(vertex)

		# Get escore for the given sequence
		escores.append(network.vs[vertex.index]["escores"])
		
		# Get sequences for all neighbors
		for neighbor in neighbors :
			sequences.append(network.vs[neighbor]["sequences"])
			escores.append(network.vs[neighbor]["escores"])

		# Initialize the empty network
		neighborhoodNet = igraph.Graph()

		# Create the neighborhood network
		
		# Create vertices
		neighborhoodNet.add_vertices(len(sequences))

		# Add attributes
		neighborhoodNet.vs["label"] = [sequences[i]+"\n"+str(escores[i]) \
										for i in range(len(sequences))]

		for i in range(1, len(sequences)) :
			neighborhoodNet.add_edge(0, i)

		# Print the list of sequences for validation purposes
		print(sequences)

		return neighborhoodNet

	# Create a network of the given repertoires where repertoire 'A' is
	# connected to repertoire 'B' if 'A' has one or more external
	# neighbors in 'B'.
	# Arguments: title=network name, genoNets=giant for the repertoire
	# TODO: We have the option of either using the GNs as nodes, or just
	#		the GN names. I need to think more about the pros and cons
	#		of each approach ...
	def createEvoNet(self, title, genoNets) :
		# Get a list of repertoire names
		repertoires = [ gn["name"].rpartition("_dominant")[0] \
						if gn["name"].rpartition("_dominant")[0]  \
						else gn["name"] \
						for gn in genoNets ]
		#repertoires = [gn["name"].replace("_dominant", "") for gn in genoNets]
		# repertoires = [gn["name"].rstrip("_dominant") for gn in genoNets]

		# Get a list of evolvability scores corresponding to the
		# repertoires
		evolvability = [gn["Evolvability"] for gn in genoNets]

		# Initialize the empty network
		network = igraph.Graph(directed=True)

		# Set title as the graph name
		network["name"] = title

		# Create vertices: Each repertoire name is a vertex in the network
		network.add_vertices(len(repertoires))

		# Add attributes

		# TODO: Remove this from the tool ...
		labels = [repertoires[i] + "\n" + str(float("{0:.2f}".format(evolvability[i]))) \
		 			for i in range(len(repertoires))]
		# labels = []
		# labels[:] = repertoires

		# Add labels to all vertices
		network.vs["label"]		= labels	# TODO: Remove this from the tool ...
		# Add repertoire names to all vertices
		network.vs["Repertoires"] = repertoires
		network.vs["GenotypeSet"] = repertoires # Redundant; added for GML
		network.vs["Evolvability"] = evolvability
	
		# Get a list of pairs of indices which represent paris of
		# repertoires that are neighbors, i.e., share at least one external
		# neighbor. The edge determination algorithm takes into account the 
		# antisymmetric relation.
		edges = [(i,j) \
					for i in range(len(repertoires)) \
					for j in range(len(repertoires)) \
					if i != j and self.repertoiresAreConnected(genoNets[i], repertoires[j])]

		# Add the above computed edges to the network
		network.add_edges(edges)

		# Add edge weights
		
		# Create a dict: keys=repertoires, values=genonets
		repToNetDict = {repertoires[i] : genoNets[i] for i in range(len(repertoires))}
		
		# Set weight for each edge, where weight=No. of evo target genotypes from i to j
		weights = [self.getEdgeWeight(edges[i], network, repToNetDict) \
									for i in range(len(edges))]

		# Add edge weights and edge labels as attributes
		network.es["weight"] = weights
		network.es["label"] = weights
		
		return network

	# Returns True if 'gn1' has at least one external neighbor in the GN for
	# 'rep2'
	def repertoiresAreConnected(self, gn1, rep2) :
		# Result flag
		areConnected = False

		# Get the list of evolvability target repertoires for gn1
		evoTargets = json.loads(gn1["Evolvability_targets"])

		# If there is at least one target in the list,
		if evoTargets :
			# If 'rep2' is in the list of targets,
			if rep2 in evoTargets :
				# 'gn1' has external neighbors in rep2
				areConnected = True

		return areConnected

	# Return the weight for the given edge of the given EN
	def getEdgeWeight(self, edge, en, repToNetDict) :
		# Get the repertoire names for source and target vertices
		srcRep = en.vs[edge[0]]["Repertoires"]
		trgRep = en.vs[edge[1]]["Repertoires"]

		# Get the source GN
		srcGN = repToNetDict[srcRep]

		# Get the list of target dicts in source GN
		allTargets = srcGN.vs["Evolvability_targets"]

		# External neighbors of source in target; to be
		# populated in the loop below.
		extNeighbors = []

		# Extract the required dict
		for trgDict in allTargets :
			# If the target repertoire is a key in this dict,
			if trgRep in trgDict :
				# Add external neighbors to the global list
				extNeighbors.extend(trgDict[trgRep])

		# Remove redundant genotypes
		extNeighbors = list(set(extNeighbors))

		return len(extNeighbors)



"""
    seq_bit_lib
    ~~~~~~~~~~~

    Library of functions for manipulation of genotypes as bits.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

from array import array


# Abstract base class that provides all methods for manipulation of arbitrary
# character sequences as bits. Only concrete classes derived from this class
# are supposed to be instantiated for use.
class AbstractBitSeqManipulator :
	# Constructor
	def __init__(self, seqLength, useIndels) :
		# Set the sequence length
		self.seqLength = seqLength

		# Flag to indicate whether shift mutations should be considered
		self.useIndels = useIndels

		# Compute and store the start indices, where each index 
		# represents the start bit of an 'n'-bit code
		self.startIndices = self.computeStartIndices()

		# Generate and store masks to be used in distance
		# computation
		self.masks = self.getMasks()

	# Basic popcount
	def popcount(self, x) :
		return bin(x).count('1')

	# Returns the 2-bit sequence corresponding to the given nucleotide letter
	def letterToBits(self, letter) :
		return self.letterToBitDict[letter]

	# Return the nucleotide letter corresponding to the given 2-bit sequence
	def bitsToLetter(self, bits) :
		return self.bitToLetterDict[bits]

	# Returns 'True' if the given number is even and 'False'
	# otherwise.
	def isEven(self, x) :
		return (x % 2) == 0

	# Returns a list of indices, such that each index in the list
	# is the index of the start of bit of the n-bit code
	# repsenting a single letter in the sequence.
	def computeStartIndices(self) :
		# Code length
		n = self.bitCodeLen
		# Total No. of bits in the sequence
		l = n * self.seqLength

		# Create a list of start indices based on the sequence
		# length
		indices = [x for x in range(n - 1, l, n)]

		return indices

	# Returns 'n' masks, where 'n = bitCodeLen'
	def getMasks(self) :
		masks = []

		for i in range(self.bitCodeLen) :
			if i == 0 :
				# Zeros mask is no use here. We start with a mask
				# with the first bit set in each letter.
				masks.append(self.generateMask(1))
			else :
				# Mask with bit i-1 set in each letter. This is
				# done by taking the i-1 power of 2.
				masks.append(self.generateMask(2 << (i - 1)))

		return masks

	# Generates and returns a mask where only every pth bit is set,
	# where 'p' is determined by 'maskVal'.
	def generateMask(self, maskVal) :
		# Total No. of bits in the sequence
		l = self.bitCodeLen * self.seqLength

		# Initialize mask to be constructed later
		mask = 0

		# Starting from the most signigicant bit, set maskval
		# in each nth position, where n=self.bitCodeLen
		for i in reversed(range(0, l, self.bitCodeLen)) :
			mask |= maskVal << i

		return mask

	# Return True if the two sequences are 1-neighbors.
	# The 1-neighborhood includes all single point mutations, as well as all 
	# left shift and right shift mutations.
	def areNeighbors(self, seq1, seq2) :
		# If the distance betweeen seq1 and seq2 is a point
		# mutation, they are neighbors; no need to check further.
		if self.distanceBwSeqs(seq1, seq2) == 1 :
			return True
		
		# If shift mutations should not be considered,
		if not self.useIndels :
			return False

		# Since seq2 is not seqarated by a point mutation, test for
		# a left shift mutation

		# Get the right most letter of seq2
		rmn = self.getLetterAtIndex(seq2, self.seqLength - 1)

		# Left shift seq1 by one letter (n bits)
		temp = seq1 << self.bitCodeLen

		# Mutate the right most nucleotide in temp and see if that results
		# in seq2. If so, seq1 and seq2 are 1-neighbors separated by
		# a single left shift mutation.		
		if seq2 == self.mutAftrLftShift(temp, self.seqLength - 1, rmn) :
			return True

		# Since seq2 is neither seqarated by a point mutation nor by
		# a left shit, test for a right shift mutation

		# Get the left most nucleotide of seq2
		lmn = self.getLetterAtIndex(seq2, 0)

		# Right shift seq1 by one nucleotide (2 bits)
		temp = seq1 >> self.bitCodeLen

		# Mutate the right most nucleotide temp and see if that results
		# in seq2. If so, seq1 and seq2 are 1-neighbors separated by
		# a single right shift mutation. If not, we have exhausted all
		# possibilites, which means seq1 and seq2 are not 1-neighbors.
		if seq2 == self.mutateLetter(temp, 0, lmn) :
			return True
		else :
			return False

	# Accepts a sequence in bit format and generates all possible 1-neighbors.
	# Returns a list of neighbors.
	def generateNeighbors(self, sequence) :
		# No. of nucleotides in the sequence
		k = self.seqLength

		# Generate all possible single letter mutants
		neighbors = [self.mutateLetter(sequence, i, target) \
						for i in range(k) \
						for target in self.bitToLetterDict.keys() \
						if target != self.getLetterAtIndex(sequence, i)]

		# If shift mutations should be considered,
		if self.useIndels :
			# Generate all possible left shift mutations

			# Left shift sequence by one letter, i.e., n bits
			temp = sequence << self.bitCodeLen

			# Generate all possible left shift mutants by mutating the right most
			# nucleotide 4 times.
			lsNeighbors = self.getShiftMutants("left", temp)
			
			# Append the list of left shift mutants to the neighborhood. Avoid duplication,
			# which is possible for sequences where the same letter occupies all positions.
			# Also, do not include the mutant that results in the source sequence.
			neighbors.extend(self.getUniqueNeighbors(lsNeighbors, neighbors, sequence))

			# Generate all right shift mutatants

			# Right shift sequence by one nucleotide (2 bits)
			temp = sequence >> self.bitCodeLen

			# Generate all possible right shift mutations by mutating the right most
			# nucleotide 4 times. However, do not include the mutant that results
			# in the source sequence.
			rsNeighbors = self.getShiftMutants("right", temp)
			
			# Append the list of right shift mutants to the neighborhood. Avoid duplication,
			# which is possible for sequences where the same letter occupies all positions.
			# Also, do not include the mutant that results in the source sequence.
			neighbors.extend(self.getUniqueNeighbors(rsNeighbors, neighbors, sequence))

		return neighbors

	# Depending on the shiftType, i.e., "left" or "right", returns the 
	# corresponding shift mutatants.
	def getShiftMutants(self, shiftType, source) :
		# Check if this is a left shift
		if shiftType == "left" :
			# The letter index at which to perform the mutations is 
			# 'seqLength - 1' for a left shift
			index = self.seqLength - 1

			# Get all left shift mutants
			mutants = [self.mutAftrLftShift(source, index, target) \
						for target in self.bitToLetterDict.keys()]
		else :	# Right shift
			# The letter index at which to perform the mutations is 
			# '0' for a left shift
			index = 0

			# Get all right shift mutants
			mutants = [self.mutateLetter(source, index, target) \
						for target in self.bitToLetterDict.keys()]

		return mutants

	# Return a list of neighbors from 'newNeighbors' that are not already in
	# 'neighbors', and none of them is the same as the 'source' sequence for
	# which the neighborhood is being generated.
	def getUniqueNeighbors(self, newNeighbors, neighbors, source) :
		return [neighbor \
				for neighbor in newNeighbors \
				if neighbor not in neighbors \
				and neighbor != source]

	# Computes the distance between two sequences (in bit format) in terms
	# of the total number of letters that differ in the two seqences.
	# FIXME: Should be extend to include distance for shift mutations ...
	def distanceBwSeqs(self, seq1, seq2) :
		# Code length
		n = self.bitCodeLen

		# Get the list of numbers that each represent diff bits for a
		# particular position between 0 and n-1.
		diffList = self.buildDiffList(seq1, seq2)

		# Since each letter is represented as an n-bit sequence, we need
		# to make sure we count the difference in a single letter only once.
		# In order to do this, within a single letter, all bits other than
		# the left most bit are shifted to the position of the left most
		# bit. This means that any differences that existed only in the 
		# other bits are now available in the left most positions as well.
		allOred = 0
		# For each number in the list
		for i in range(len(diffList)) :
			# Move all differences to the left most position
			allOred |= diffList[i] << (n - i)

		# Popcount gives the distance
		distance = self.popcount(allOred)

		return distance

	# Builds a list of numbers, where each number has bits set only at
	# specific indices, and represents difference in bit value between
	# seq1 and seq2 at that index. These indices appear at intervals of 
	# length 'n', where 'n = bitCodeLen'. Each number repesents the bits 
	# for each letter that differs on a certain index within the bit code. 
	# E.g., one of the numbers contains only the difference bits on the first
	# index of each letter.
	def buildDiffList(self, seq1, seq2) :
		diffList = []

		# Get the difference values for all bits. XOR gives us only those 
		# bits that have different values in the two sequences.
		diffBits = seq1 ^ seq2	# Bitwise XOR

		# Populate the diffLists corresponding to each position
		for i in range(self.bitCodeLen) :
			diffList.append(diffBits & self.masks[i])

		return diffList

	# Assuming the requested single letter mutation is being performed
	# on a sequence that has been shifted left, returns a mutant with the
	# letter at index 'position' in 'sequence' replaced with 'target'.
	# Note: This function is needed because for sequences that start with
	# 		a letter other than '0', a left shift can leave one or more 
	#		'1' bits on the left. These must be unset so that integer 
	#		comparisons return correct results.
	def mutAftrLftShift(self, sequence, position, target) :
		# Mutate the required nucleotide
		mutant = self.mutateLetter(sequence, position, target)

		# Unset any possible set bits on the left at position -1
		mutant = self.mutateLetter(mutant, -1, 0)

		return mutant

	# Mutates a single letter in the bit sequence at index=position, to
	# value=target. Position is expected to have been determined using 
	# zero-based indexing.
	def mutateLetter(self, sequence, position, target) :
		# Code length
		n = self.bitCodeLen
		# Total No. of bits in the sequence
		l = n * self.seqLength

		# Get a mask that is all ones of length=n
		ones = (1 << n) - 1

		# Get index of the right bit in the n-bit sequence that
		# corresponds to the given position.
		bitIndex = (l - n) - (n * position)
		
		# Create an n-bit mask, i.e., all ones ('11...1') at the 
		# index computed above
		mask = ones << bitIndex

		# Perform the mutation
		# TODO: Add desciption here ...
		mutatedSeq = sequence & (~mask) | (target << bitIndex)

		return mutatedSeq

	# Return the n-bit code at the given index, where index is in the 
	# range '0' to 'k - 1', i.e, it is the index corresponding to the
	# letter index in the string sequence (inreasing from left to right).
	def getLetterAtIndex(self, sequence, index) :
		# Total No. of bits in the sequence
		l = self.bitCodeLen * self.seqLength

		# Calculate the index that corresponds to the to the start bit
		# for the letter on the given index. E.g., if k=8, n=5, l=40,
		# index=7: l-1=39, index*n=35, => i=4. Then the letter is
		# represented by bits on indices 4,3,2,1,0.
		i = (l - 1) - (index * self.bitCodeLen)

		# Get the letter on index 'i'
		letter = self.getbitsOnIndex(i, sequence)

		return letter

	# Convert a sequence into a bit representation
	def seqToBits(self, sequence) :
		# Initializing the to-be-constructed bit sequence
		bitSequence = 0

		# For each letter in the sequence
		for i in range(0, self.seqLength) :
			# Get the corresponding bit code
			bitValue = self.letterToBitDict[sequence[i]]

			# Add the bit code to the bit sequence
			bitSequence |= bitValue

			# If this is not the last letter in the sequence
			if i != (self.seqLength - 1) :
				# Left shift by 'bitCodeLen' bits to make room 
				# for the bit code corresponding to the next 
				# letter in the sequence
				bitSequence <<= self.bitCodeLen

		return bitSequence

	# Convert a bit sequence into letter sequence	
	def bitsToSeq(self, bitSequence) :
		# k-mer length
		k = self.seqLength
		# Code length
		n = self.bitCodeLen

		# String to store the sequence. The initialization here has two
		# purposes: 1) To get the desired size, 2) To ensure there are
		# no ones, since left shifting ones at the left most index can
		# cause problems.
		outputSequence = array('c', ["0" for i in range(k)])

		# For each n-bit sequence (in a k-mer occupying n*k bits in an 
		# integer),
		j = k - 1
		for i in self.startIndices :
			# Get the letter on index 'i'
			letter = self.getbitsOnIndex(i, bitSequence)

			# Convert the n-bit sequence into a letter and store
			# it in the correct index of the output sequence
			outputSequence[j] = self.bitToLetterDict[letter]

			j = j - 1

		return outputSequence.tostring()

	# Extracts 'n' bits from 'bitSeq' starting at index 'i'.
	# Returns an integer that is a copy of only these bits.
	def getbitsOnIndex(self, i, bitSeq) :
		# Code length
		n = self.bitCodeLen

		# Extract the ith bit
		letter = ((bitSeq & (1 << i)) >> i)

		# For the rest of the bits used to represent this letter
		for p in range(1, n) :
			# Shift the bit by 1 bit to the left to make room
			letter <<= 1

			# Extract pth bit for the current letter
			letter |= ((bitSeq & (1 << (i - p))) >> (i - p))

		return letter




"""
    peaks_functions
    ~~~~~~~~~~~~~~~

    Contains functions used for computation of peaks.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

from genonets_utils import Utils


# Provides functions to analyze the landscape defined by the geontype
# network, and detect peaks in the landscape.
# TODO: Once the peaks dict has been created for a network, perhaps the
#		path list can be deleted to release memeory ...
class PeakAnalyzer :
	# Constructor
	def __init__(self, network, netUtils, delta) :
		# Reference to the network on which to perform this
		# analysis
		self.network = network

		# Get a reference to the NetworkUtils object
		self.netUtils = netUtils

		# Get a reference to the BitSeqManipulator in use
		self.bitManip = netUtils.bitManip

		# Keep a copy of the delta value
		self.delta = delta

		# Set once peaks have been identified
		self.peaks = None

	# Returns the peak dict which has seq as one of the keys
	def getPeakWithSeq(self, seq) :
		# Get all peaks
		peaks = self.getPeaks()

		# For each peak
		for peakId in peaks.keys() :
			# Get the peak dict
			peak = peaks[peakId]

			# Check if the given seq is in the peak
			if seq in peak["sequences"] :
				return peak

		# TODO: replace with proper exception handling ...
		print("Error: Count not find " + seq + " in any peak!!!!")
		return None

	# The function to be exposed to the calling object. Triggers the
	# peak detection algorithm and returns the result as a dictionary
	# with peak Ids as keys.
	def getPeaks(self, recompute=False) :
		# If peak computation has not been done, or the caller has
		# explicitly asked for re-running the algorithms,
		if not self.peaks or recompute :
			# Get a structured array with tuples (sequence, escore),
			# sorted in descending order of e-scores.
			sortedArr = Utils.getSortedSeqEscArr(self.network, 
								self.bitManip.seqLength,
								sortOrder="descending")

			# Identify peaks
			self.peaks = self.buildPeaks(sortedArr)

		return self.peaks

	# Orchistration function for the peak detection algorithm.
	def buildPeaks(self, elements) :		
		peaks = {}

		# List of elements that have been processed independently, or as
		# part of a neighborhood, e.g., all elements in a plateau.
		processed = []

		# Process each element, where elements are sorted in the
		# descending order of e-scores.
		for i in range(len(elements)) :
			# If the element is in the list of already processed elements,
			if elements[i] in processed :
				# Skip iteration
				continue

			# Build a list of elements that belong to the neutral
			# zone, i.e., build the bin/band.
			binMembers = self.getBin(i, elements)

			# Build a list of elements within the bin that are
			# connected. This requires a BFS. If the iteration element
			# turns out to be a peak, all elements in this list will
			# be part of the plateau. In addition, get a list of elements
			# that are in the bin, but not connected.
			neutralZone, nonNeighs = self.getNeutralZone(elements[i], binMembers)
			# Append the element itself as well
			neutralZone.append(elements[i])

			# Get a list of peaks for which the focal element falls within
			# the bin.
			peakNonNeighs = self.getPeaksNonNeighs(elements[i], peaks)

			# Check connectivity with existing peaks. If the focal element
			# is connected to a peak, and it does not lie within the bin for
			# that peak, the entire neutral zone is a 'non-peak'.
			neighboringPeaks = self.getNeighPeaks(neutralZone, peakNonNeighs, peaks)

			# Depending on the results so far, append the list to an
			# existing peak, or create a new peak.
			if neighboringPeaks :
				# Append the entire neutral zone to the path list for the 
				# neighboring peaks.
				self.appendToPeaks(neutralZone, neighboringPeaks, peaks)
				
				# Append the neutral zone to the processed list. If an entire
				# zone has been detected as connected to a peak, none of the
				# sequences can be peaks, and therefore do not need to
				# be processed independently.
				if len(neutralZone) > 1 :
					processed.extend(neutralZone)
			else :
				# Create a new peak, and add all members of the neutral zone
				# to this peak.
				peakId = len(peaks)
				peaks[peakId] = self.createPeak(neutralZone, nonNeighs)

				# If the peak is a plateau, mark all elements as processed,
				# so that none of them is processed independently.
				if len(neutralZone) > 1 :
					processed.extend(neutralZone)

		return peaks

	# Returns a list of elements that fall within the noise determined score
	# zone for the given focal element, identified by the given index.
	def getBin(self, index, elements) :
		elBin = []

		focalElement = elements[index]

		# Starting from the next element in the list and until the
		# end
		for i in range(index + 1, len(elements)) :
			# If the escore is within the band
			if elements[i]['escore'] >= focalElement['escore'] - self.delta :
				elBin.append(elements[i])
			else :
				# Once we've reached an element which lies in a lower
				# band, we know that the remaining elements will also
				# lie in lower bands. Therefore, we don't need to look
				# any further.
				break

		return elBin

	# The neutral zone consists of all elements that satisfy the following 
	# conditions:
	# 1) The element lies within the bin for the given focal element
	# 2) The element is either a 1-neighbor of the focal element, or
	# 3) The element is an 'k'-neighbor of the focal element, such that all
	#	 elements that constitute the path between the element and the focal
	#	 element lie within the bin.
	# Returns the neutral zone, as well as a list of elements that are within
	# the bin, but are not connected to the focal element.
	def getNeutralZone(self, focalElement, binMembers) :
		# Get the list of sequences that constitute the neighborhood of the
		# focal element
		neighSeqs = self.getNeighSeqs(focalElement)

		# Get a list of elements that exist in the neighborhood of the
		# focal element. These are by definition already part of the neutral
		# zone.
		neutralZone = [item for item in binMembers if item['sequence'] in neighSeqs]

		# Get all non-neighbor elements in the bin
		nonNeighs = [binMembers[i] for i in range(len(binMembers)) \
						if binMembers[i] not in neutralZone]

		# Do a breadth-first search of the neutral zone to find elements in
		# the non-neighbors list that might be indirected connected.
		neutralZone, nonNeighs = self.bfsNeutralZone(neutralZone, nonNeighs)

		return neutralZone, nonNeighs

	# Performs a breadth-first search of the neutral zone to check if any of
	# the elements in the list of non-neighboring bin memebers are connected
	# to any of the members in the neutral zone. The search is dynamic, i.e.,
	# as connected elements are found, these are themselves then considered
	# as members of the neutral zone, and used as seed elements for further
	# search.
	# Returns the complete neutral zone, as well as a list of non-neighbors
	# left after the BFS.
	def bfsNeutralZone(self, neutralZone, nonNeighs) :		
		i = 0	# Index variable

		# Keep iterating for as long as there are elements left in the
		# neutral zone.
		while i < len(neutralZone) :
			removeList = []

			# For each non-neighbor
			for nonNeigh in nonNeighs :
				# If the non-neighbor is connected to the ith element in the
				# neutral zone,
				if self.netUtils.areConnected(nonNeigh['sequence'], \
						neutralZone[i]['sequence']) :
					# Append the non-neighbor to the neutral zone
					neutralZone.append(nonNeigh)
					
					# Mark the non-neighbor, so that it is removed from the
					# non-neighbors list after the for loop.
					removeList.append(nonNeigh)

			# Remove the elements from the non-neighbors list that have already
			# been added to the neutral zone.
			nonNeighs = [item for item in nonNeighs if item not in removeList]

			# Move on to the next element in the neutral zone
			i += 1

		return neutralZone, nonNeighs

	# Checks if the given element was identified as a non-neighbor in the
	# bin for an existing peak.
	def getPeaksNonNeighs(self, focalElement, peaks) :
		# Note: We do not check whether or not the e-score of the focal
		#		element + delta >= e-score of the peak summit, because
		#		if it weren't, the focal element here would not have
		#		made it to the non-neighbors list.
		return [peakId for peakId in peaks.keys() \
					if focalElement in peaks[peakId]["non-neighbors"]]

	# For each element in the neutral zone, checks whether it is connected
	# to an existing peak, except for peaks that are in 'peakNonNeighs'.
	# Returns a list of 'peakIds' for the neighboring peaks if any.
	def getNeighPeaks(self, neutralZone, peakNonNeighs, peaks) :
		# List of peak Ids of all peaks to which this neutral zone is
		# connected.
		neighboringPeaks = []

		# For each element in the neutral zone
		for element in neutralZone :
			# Get peaks Id of all neighboring peaks for this element
			peaksAsNeighs = self.getNeighPeaksFor(element, peakNonNeighs, peaks)

			# The list should be empty if no existing peaks neighbor this
			# element
			if peaksAsNeighs :				
				neighboringPeaks.extend(peaksAsNeighs)

		# Remove repeated values
		neighboringPeaks = list(set(neighboringPeaks))

		return neighboringPeaks
	
	# Checks if the given element is connected to a peak that is not in
	# 'peakNonNeighs'.
	# Returns a list of 'peakIds' for the neighboring peaks if any.
	def getNeighPeaksFor(self, element, peakNonNeighs, peaks) :
		peakIds = []

		# For each existing peak,
		for peakId in peaks.keys() :
			# If the focal element lies within the bin for this peak,
			if peakId in peakNonNeighs :
				# No need to check for indirect connectivity. Move on
				# to the next peak.
				continue

			# Get the list of connected elelments for this peak
			pathList = peaks[peakId]["list"]

			# For each element in the path list,
			for item in pathList :
				# Check if the given element is connected to current item
				# in the path list. The assumption here is that the first 
				# items in the list are the peak sequences themselves.
				if self.netUtils.areConnected(item['sequence'], element['sequence']) :
					# The neutral zone has a connected to the peak. Therefore,
					# add the peak ID to the list of connected peaks.
					peakIds.append(peakId)

					# Move on to the next peak
					break

		return peakIds

	# Creates a peak which comprises the given neutral zone. Also, stores
	# the list of unconncted bin members.
	def createPeak(self, neutralZone, nonNeighs) :
		# Dictionary: There are two keys, 1) List of sequences that
		# constitute the peak, and 2) list of sequences that are
		# connected to the peak.
		peak = {}

		# List sequences that constitute the peak. Reverse the neutral zone
		# so that the focal element that was appended at the end becomes 
		# the head.
		neutralZone.reverse()
		peak["sequences"] = [neutralZone[i][0] \
								for i in range(len(neutralZone))]

		# Initialize the list of the connected sequences with sequences
		# in the peak. This simplifies comparisons when searching through
		# the list for connectivity.		
		peak["list"] = neutralZone

		# Store the non-neighboring elements from the same band to simplify
		# lookup operation during peak determination.
		peak["non-neighbors"] = nonNeighs
		
		return peak

	# Appends the given neutral zone to all the given neighboring peaks.
	def appendToPeaks(self, neutralZone, neighboringPeaks, peaks) :
		# For each neighboring peak
		for peakId in neighboringPeaks :
			# Append the neutral zone to the list of elements found to
			# be connected to the peak.
			peaks[peakId]["list"].extend(neutralZone)

	# Returns the list of sequences that constitute the neighborhood of the
	# given element
	def getNeighSeqs(self, element) :
		return [self.network.vs[neighbor]["sequences"] \
					for neighbor in self.netUtils.getNeighbors( \
						element['sequence'], self.network)]




"""
    overlap_functions
    ~~~~~~~~~~~~~~~~~

    Contains functions used for computation of overlap.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

class OverlapAnalyzer :
	# Constructor
	def __init__(self, repToGiantDict, repertoires) :
		# Copy of dict: key=repertoire, value=giant
		self.repToGiantDict = repToGiantDict

		# List of available repertoires
		self.repertoires = repertoires

	def getOverlapData(self) :
		if len(self.repertoires) < 2 :
			print("Overlap computation triggered with only one repertoire!")
			print("Overlap can only be calculated with 2 or more repertoires.")

			return None, None

		# Overlap dict for all repertoires. Dict{rep : {seq : [target reps]}}
		allOverlap = { rep : { seq : [] \
								for seq in self.repToGiantDict[rep].vs["sequences"] } \
						for rep in self.repertoires }

		# Initialize the overlap matrix with zeros
		overlapMat = [ [0 for x in range(len(self.repertoires))] \
							for x in range(len(self.repertoires)) ]

		# For each repertoire,
		for i in range(len(self.repertoires) - 1) :
			# Get giant for this repertoire
			giantI = self.repToGiantDict[self.repertoires[i]]

			# Get sequence list from giant
			seqsI = giantI.vs["sequences"]

			# For the rest of the repertoires,
			for j in range(i+1, len(self.repertoires)) :
				# Get giant for this repertoire
				giantJ = self.repToGiantDict[self.repertoires[j]]

				# Get sequence list from giant
				seqsJ = giantJ.vs["sequences"]

				# Get the list of sequences that are common to both repertoires
				overlapList = self.getOverlapList(seqsI, seqsJ)

				# Get the number of sequences that overlap between the
				# two repertoires, and set it in the matrix
				overlapMat[i][j] = overlapMat[j][i] = len(overlapList)

				# If there is at least one shared sequence,
				if len(overlapList) > 0 :
					# For each sequence in the list,
					for sequence in overlapList :
						# Append the target to the target list for this sequence
						allOverlap[self.repertoires[i]][sequence].append(self.repertoires[j])
						allOverlap[self.repertoires[j]][sequence].append(self.repertoires[i])

		return overlapMat, self.repertoires, allOverlap

	# Returns a list of sequences that are common to the two input sequence
	# lists
	def getOverlapList(self, seqs1, seqs2) :
		return list(set(seqs1) & set(seqs2))




"""
    genonets_exceptions
    ~~~~~~~~~~~~~~~~~~~

    Defines custom exceptions.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

class GenonetsError(Exception) :
	def __init__(self, errId, info="") :
		self.errId = errId
		self.info = info

	def __str__(self) :
		return repr(self.errId)




"""
    genonets_utils
    ~~~~~~~~~~~~~~

    General utility functions.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import numpy as np


class Utils :
	# Reverses the given dictionary, i.e., sets keys as values and
	# vice versa. Returns the reverse dictionary.
	# Note: This can only work if the values in the given dict are
	#		hashable. This function will not work for lists, etc.
	#		Keys and values must have a one-to-one mapping.
	@staticmethod
	def reverseDict(inDict) :
		return {inDict[key]: key for key in inDict.keys()}

	@staticmethod
	def getSeqWithMaxScore(network, seqLength) :
		# Get structured array with tuples: (sequence, escore),
		# sorted in ascending order of escores.
		sortedArr = Utils.getSortedSeqEscArr(network, seqLength, 
						sortOrder="ascending")

		# Last sequence in the array is the one with the highest
		# e-score.
		return sortedArr[-1]['sequence']

	# Returns a structured array with tuples: (sequence, escore),
	# sorted in either 'ascending' or 'descending' order of escores;
	# as determined by the 'sortOrder' received as argument.
	@staticmethod
	def getSortedSeqEscArr(network, seqLength, sortOrder) :
		# Get the structred array
		seqEscrArr = Utils.getSeqEscrArr(network, seqLength)

		# Sort the array in ascending order of escores
		sortedArr = np.sort(seqEscrArr, order="escore")

		# If the required order is 'ascending',
		if sortOrder == "ascending" :
			# We already have what we need
			return sortedArr
		
		# If the required order is 'descending',
		if sortOrder == "descending" :
			# Get a reverse view, which is equivalent of sorting
			# in descending order.
			revArr = sortedArr[::-1]

			return revArr

	# Generates a structured array with tuples: (sequence, escore) for 
	# all genotypes in the network.
	@staticmethod
	def getSeqEscrArr(network, seqLength) :
		# Get a list of all sequences in the network
		seqs = network.vs["sequences"]

		# Get a list of all escores
		escores = network.vs["escores"]

		# Create the structured array
		dtype = [('sequence', 'S' + str(seqLength)), 
					('escore', float)]
		values = [(seqs[i], escores[i]) for i in range(len(seqs))]

		return np.array(values, dtype=dtype)




"""
    path_functions
    ~~~~~~~~~~~~~~

    Contains functions used for computation of accessible mutational paths.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import igraph

#from memory_profiler import profile

from genonets_utils import Utils


class PathAnalyzer :
	# Constructor
	def __init__(self, network, netUtils, delta) :
		# Reference to the network on which to perform this
		# analysis
		self.network = network

		# Get a reference to the NetworkUtils object
		self.netUtils = netUtils

		# Get a reference to the BitSeqManipulator in use
		self.bitManip = netUtils.bitManip

		# Keep a copy of the delta value
		self.delta = delta

		# Summit vertex, to be populated later.
		self.summitId = None

		# Dict to store all accessible paths. {vertexId : [paths]}
		self.allPathsToPeak = self.initPathsToPeak()

	def initPathsToPeak(self) :
		return { vId : [] for vId in range(len(self.network.vs["sequences"])) }

	def getSummitId(self) :
		return self.summitId

	def getAllPathsToPeak(self) :
		return self.allPathsToPeak

	def getPathsThruVtxs(self) :
		# List: Each element is the No. of paths through the vertex Id
		# 		corresponding to the index
		pathsThruVtx = [ 0 for i in range(self.network.vcount()) ]

		# Go through accessible paths corresponding to each vertex
		for vtxId in range(self.network.vcount()) :
			# Get the list of accessible paths for this vertex
			vtxPaths = self.network.vs[vtxId]["pathsToSummit"]

			# For each path,
			for path in vtxPaths :
				# For each vertex in the path,
				for vtx in path :
					# Increment the No. of paths that go through
					# this vertex
					pathsThruVtx[vtx] += 1

		return pathsThruVtx

	# Get the ratio of accessible paths to all paths of the
	# given length.
	def getAccessiblePaths(self, pathLength) :
		# Stats for the entire network
		totalPaths  = 0	# Total mutational paths (only shortest)
		allAccPaths = 0	# No. of accessible paths (only shortest)

		# Get the sequence with the highest score. This sequence
		# will represent the global peak. All paths use this
		# sequence as the target.
		summit = Utils.getSeqWithMaxScore(self.network, self.bitManip.seqLength)

		# Get the vertex object that represents the summit
		trgtVrtx = self.netUtils.getVertex(summit, self.network)

		# Store a copy of the summit
		self.summitId = trgtVrtx.index

		# Get a list of all sequences in the network
		sequences = self.network.vs["sequences"]
		# Remove the target itself
		sequences.remove(summit)

		# For each sequence in the network
		for source in sequences :
			# Get all shortest paths as well as all accessible
			# shortest paths from source to summit
			self.getShortestAccPaths(source, trgtVrtx, 0)

			# shrtPaths, accPaths = self.getShortestAccPaths(source, trgtVrtx, 
			# 							pathLength + 1)

		# 	# If at least one shortest path of length == 'pathLength' was
		# 	# found,
		# 	if shrtPaths :
		# 		# Increment the counts
		# 		totalPaths  += float(len(shrtPaths))
		# 		allAccPaths += float(len(accPaths))

		# try :
		# 	return float(allAccPaths) / float(totalPaths)
		# except ZeroDivisionError :
		# 	return 0

		return

	# From within all shortest paths between source sequence and the 
	# given target vertex, computes all accessible paths.
	# For computation, returns only those paths that are of the given
	# length. However, all accessible paths are stored regardless of
	# size. This is then used in the visualization.
	#@profile
	def getShortestAccPaths(self, source, trgtVrtx, pathLength) :
		# Get the source and target vertices
		srcVrtx  = self.netUtils.getVertex(source, self.network)

		# Get all shortest paths between source and target
		allShrtPaths = self.network.get_all_shortest_paths(srcVrtx, 
			trgtVrtx, mode=igraph.OUT)

		# Get all shortest accessible paths
		shrtAccPaths = [ self.network.vs[path].indices \
							for path in allShrtPaths \
							if self.isAccessible(path) ]
		# Store the paths
		self.allPathsToPeak[srcVrtx.index].extend(shrtAccPaths)

		# # If the shortest path length is the same as the required
		# # length,
		# if len(allShrtPaths[0]) == pathLength :
		# 	allShrtAccPaths = [ self.network.vs[path]["sequences"] \
		# 						for path in allShrtPaths \
		# 						if self.isAccessible(path) ]
		# else :
		# 	# We did not find a path of the required length
		# 	allShrtPaths = None
		# 	allShrtAccPaths = None

		# return allShrtPaths, allShrtAccPaths

		return

	# Determines whether the given path is an accessible path, i.e.,
	# scores on this path increase monotonously.
	def isAccessible(self, path) :
		isAcc = True

		# Get a list of escores for the sequences in the path
		escores = self.network.vs[path]["escores"]

		# Place holder to keep track of the highest score
		# encountered inside the loop
		maxYet = -0.5
		# For each escore in the list
		for i in range(len(escores) - 1) :
			# If the current score is higher than the max so far,
			if escores[i] > maxYet :
				# Assign current escore to max
				maxYet = escores[i]

			# If the next sequence in the path is in a lower bin,
			if (maxYet - self.delta > escores[i + 1]) :
				# The increase in e-score is not monotonous. Hence,
				# the path is not accessible.
				isAcc = False
				break

		return isAcc




"""
    genonets_writer
    ~~~~~~~~~~~~~~~

    Helper functions for writing results to files.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import os
import copy
import warnings

from genonets_constants import ErrorCodes
from genonets_exceptions import GenonetsError
from genonets_constants import GenonetsConstants as gc

class Writer :
	@staticmethod
	def writeInParamsToFile(paramsDict, path) :
		fileName = path + "in_params.txt"

		# If the required directories have not already been created
		if not os.path.exists(os.path.dirname(path)):
			# Create the directories
			try :
				os.makedirs(os.path.dirname(path))
			except os.error :
				print("Error: " + 
						ErrorCodes.getErrDescription(ErrorCodes.CANNOT_CREATE_DIRECTORY) +
							": Path - " + path)
								
				raise GenonetsError(ErrorCodes.CANNOT_CREATE_DIRECTORY,
					"Path - " + path)

		# Open the file
		try :
			with open(fileName, "w") as outFile :
				# For each input parameter,
				for param in paramsDict.keys() :
					outFile.write(param + ": " + paramsDict[param] + "\n")
		except Exception :
			print("Error: " + 
					ErrorCodes.getErrDescription(ErrorCodes.CANNOT_WRITE_TO_FILE) +
						": Path - " + path)
			
			raise GenonetsError(ErrorCodes.CANNOT_WRITE_TO_FILE)

	@staticmethod
	def writeNetsToFile(repToNetDict, repToGiantDict, netBuilder, path,  attrsToIgnore, repertoires=gc.ALL) :
		# If all repertoires should be considered,
		if repertoires == gc.ALL :
			# Get a list of all repertoires
			repertoires = repToNetDict.keys()

		# If the required directories have not already been created
		if not os.path.exists(os.path.dirname(path)):
			# Create the directories
			os.makedirs(os.path.dirname(path))

		# For each repertoire
		for repertoire in repertoires :
			# Write the entire network to file in GML format
			Writer.writeNetToFile(repToNetDict[repertoire], path, attrsToIgnore)

			# Get the number of components in the network
			numComponents = len(netBuilder.getComponents(repToNetDict[repertoire]))

			# If the network as more than one components,
			if numComponents > 1 :
				#  Write the giant component to the file in GML format
				Writer.writeNetToFile(repToGiantDict[repertoire], path, attrsToIgnore)

	@staticmethod
	def writeNetToFile(network, path, attrsToIgnore) :
		# File name		
		fileName = path + network["name"] + ".gml"

		# Create a deep copy of the network
		netToWrite = copy.deepcopy(network)

		# Get a list of network attributes
		netAttrs = netToWrite.attributes()

		# Remove the network attributes that need not be written to file
		for attr in attrsToIgnore(level="network") :
			if attr in netAttrs :
				del netToWrite[attr]

		# Get a list of vertex attributes
		vtxAttrs = netToWrite.vs.attributes()

		# Remove the vertex attributes that need not be written to file
		for attr in attrsToIgnore(level="vertex") :
			if attr in vtxAttrs :
				del netToWrite.vs[attr]

		# If the given network is a genotype network, and not a
		# phenotype network,
		if not network.is_directed() :
			# Rename 'sequences' and 'escores'
			netToWrite.vs["genotype"] = netToWrite.vs["sequences"]
			netToWrite.vs["score"] = netToWrite.vs["escores"]
			del netToWrite.vs["sequences"]
			del netToWrite.vs["escores"]

		# Write network to file in GML format, while ignoring 'igraph'
		# related warnings.
		with warnings.catch_warnings() :
			# Create a filter for 'igraph' warnings
			warnings.simplefilter("ignore")
			
			# Write GML file
			netToWrite.write(fileName, format="gml")

	@staticmethod
	def writeNetAttribs(repToNetDict, repToGiantDict, netBuilder, path, attrsToIgnore, repertoires=gc.ALL) :
		# If all repertoires should be considered,
		if repertoires == gc.ALL :
			# Get a list of all repertoires
			repertoires = repToNetDict.keys()

		# Construct the list of attributes to be written to file
		attributes = repToNetDict[repertoires[0]].attributes()
		attributes.remove("name")

		# Get the number of components in the network
		numComponents = len(netBuilder.getComponents(repToNetDict[repertoires[0]]))

		# If the network as more than one components,
		if numComponents > 1 :
			# Get attributes from giant as well
			attributes.extend(repToGiantDict[repertoires[0]].attributes())

		# Remove the attributes that need not be written to file
		for attr in attrsToIgnore() :
			if attr in attributes :
				attributes.remove(attr)

		# If the required directories have not already been created
		if not os.path.exists(os.path.dirname(path)):
			# Create the directories
			os.makedirs(os.path.dirname(path))

		fileName = path + "Genotype_set_measures.txt"
		dataFile = open(fileName, 'w')

		# Write the row of headers
		dataFile.write("Genotype_set" + "\t")
		for attribute in attributes :
			dataFile.write(attribute + "\t")

		dataFile.write("\n")

		# For each repertoire,
		for repertoire in repertoires :
			# Write the network name
			dataFile.write(repToNetDict[repertoire]["name"] + "\t")
			# For each attribute,
			for attribute in attributes :
				# If the attribute is in the main network,
				if attribute in repToNetDict[repertoire].attributes() :
					dataFile.write(str(repToNetDict[repertoire][attribute]) + "\t")
				else :
					dataFile.write(str(repToGiantDict[repertoire][attribute]) + "\t")

			dataFile.write("\n")

		dataFile.close()

	@staticmethod
	def writeSeqAttribs(repToNetDict, repToGiantDict, netBuilder, path, attrsToIgnore, repertoires=gc.ALL) :
		# If all repertoires should be considered,
		if repertoires == gc.ALL :
			# Get a list of all repertoires
			repertoires = repToGiantDict.keys()

		# If the required directories have not already been created
		if not os.path.exists(os.path.dirname(path)):
			# Create the directories
			os.makedirs(os.path.dirname(path))

		for repertoire in repertoires :
			# Create the file name
			fileName = path + repertoire + "_genotype_measures.txt"

			# Open file to write
			with open(fileName, 'w') as dataFile :
				Writer.writeSeqAttribsFor(repertoire, repToGiantDict[repertoire], 
					dataFile, attrsToIgnore)

	@staticmethod
	def writeSeqAttribsFor(repertoire, network, dataFile, attrsToIgnore) :
		# Get vertex level attributes
		attributes = network.vs.attributes()

		# Remove the attributes that need not be written to file
		for attr in attrsToIgnore() :
			if attr in attributes :
				attributes.remove(attr)

		# Write the row of headers
		dataFile.write("Sequence" + "\t")
		for attribute in attributes :
			dataFile.write(attribute + "\t")

		dataFile.write("\n")

		sequences = network.vs["sequences"]

		# For each sequence,
		for i in range(len(sequences)) :		
			# Write the sequence
			dataFile.write(sequences[i] + "\t")
			# For each attribute,
			for attribute in attributes :
				# Write value to file
				dataFile.write(str(network.vs[i][attribute]) + "\t")

			dataFile.write("\n")

	@staticmethod
	def writeOverlapToFile(overlapMat, repertoires, path) :
		fileName = path + "Genotype_set_overlap.txt"

		# If the required directories have not already been created
		if not os.path.exists(os.path.dirname(path)):
			# Create the directories
			os.makedirs(os.path.dirname(path))

		dataFile = open(fileName, 'w')

		# Write header: All repertoire names as column headers
		for repertoire in repertoires :
			dataFile.write("\t" + repertoire)

		# Write the matrix with each row beginning with the
		# corresponding repertoire name
		for i in range(len(overlapMat)) :
			dataFile.write("\n")

			# Write RBP name
			dataFile.write(repertoires[i])

			for j in range(len(overlapMat)) :
				if not i == j :
					dataFile.write("\t" + str(overlapMat[i][j]))
				else :
					#dataFile.write("\t" + "######")
					dataFile.write("\t" + "NaN")

		dataFile.close()




"""
    evolvability_functions
    ~~~~~~~~~~~~~~~~~~~~~~

    Contains functions used for genotype and phenotype evolvability computations.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

class EvolvabilityAnalyzer :
	# Constructor
	def __init__(self, network, dataDict, seqToRepDict, repToGiantDict, netBuilder, bitsToSeqDict) :
		# Reference to the network on which to perform this
		# analysis
		self.network = network

		# Reference to dict: key=rep, value=dict{seq: score}
		self.dataDict = dataDict

		# Copy dict: key=sequence, value=[repertoires]
		self.seqToRepDict = seqToRepDict.copy()

		# Copy of dict: key=repertoire, value=giant
		self.repToGiantDict = repToGiantDict

		# Remove all seq-rep mappings, where seq is not within the
		# giant
		self.updateSeqToRepDict()

		# Get a reference to the NetworkUtils object
		self.netBuilder = netBuilder

		# Dict {bitseq : seq}
		self.bitsToSeqDict = bitsToSeqDict

		# Reference to the list of evlvability values for all sequences
		self.evoTuples = None

	def updateSeqToRepDict(self) :
		# For each sequence key in the dict,
		for seq in self.seqToRepDict.keys() :
			# For each repertoire of which the sequence is a member,
			for rep in self.seqToRepDict[seq] :
				# If the sequence is not within the giant of the
				# repertoire,
				if seq not in self.repToGiantDict[rep].vs["sequences"] :
					# Remove the repertoire from the list
					self.seqToRepDict[seq].remove(rep)

	# Returns repertorire evolvability
	def getReportoireEvo(self) :
		# Get evo tuples for all sequences in the network
		evoTuples = self.getEvoAll()

		# Get repertoire lists from the tuples
		repLists = [evoTuples[1].keys() \
						for evoTuples in evoTuples if evoTuples[1]]

		# Combine all repLists into one list
		targets = []
		for repList in repLists :
			targets.extend([target for target in repList])

		# Remove redundant repertoires
		targets = list(set(targets))

		try :
			evolvability = float(len(targets)) / float(len(self.dataDict) - 1)

			return evolvability, targets			
		except ZeroDivisionError :
			return 0, targets
	
	def getEvoAll(self, recompute=False) :
		# If either the evolvability values have not been computed already,
		# or the caller has explicitly requested re-computing,
		if not self.evoTuples or recompute :			
			# Get tuples: (evo, targetRepsDict) for all sequences in the network
			self.evoTuples = [self.getSeqEvo(seq) \
								for seq in self.network.vs["sequences"]]

		return self.evoTuples

	# Returns evolvability score for the given sequence
	def getSeqEvo(self, sequence) :
		# Get a list of sequences that are 1-neighbors, but not part of the
		# genotype network
		externNeighbors = self.netBuilder.getExternalNeighbors(sequence, self.network)

		# Get dictionary: key=repertoire, value=extNeighs. The keys represents the
		# list of repertoires to which the given sequence could join by undergoing 
		# a single mutation.
		targetReps = self.getEvoTargetReps(sequence, externNeighbors)

		# Compute and return the evolvability score, i.e., No. of repertoire
		# targets found for this sequence / No. of repertoires available in
		# the dataset.
		# FIXME: ZeroDivisionError can only occur if the denominator is zero. This
		#		 already be checked at the very beginning of the function to avoid
		#		 the rest of the computation ...
		try :
			evolvability = float(len(targetReps)) / float(len(self.dataDict) - 1)
		except ZeroDivisionError :
			evolvability = 0

		return (evolvability, targetReps)

	# For the given sequence, list of external neighbors, and the sequence
	# to repertoire dictionary, return a list of repertoires for which the given 
	# sequence could develop score >= tau by a single mutation.
	def getEvoTargetReps(self, sequence, extNeighs) :
		targetReps = {}

		# For each external neighbor,
		for extNeigh in extNeighs :
			try :
				# Get the sequence corresponding to the bitseq
				extNeighSeq = self.bitsToSeqDict[extNeigh]
			except KeyError :
				# If the bitseq corresponding to sequence could not be found,
				# it means that it does not exist in any other repertoire. We
				# can therefore safely skip the current iteration
				continue

			# If the external neighbor is in any repertoires,
			if extNeighSeq in self.seqToRepDict :
				# For each repertoire that contains the external neighbor,
				for repertoire in self.seqToRepDict[extNeighSeq] :
					# We have found a repertoire with a genotype network that
					# has a 1-neighbor to which  the sequence can evolve.
					if repertoire not in targetReps :
						targetReps[repertoire] = []

					targetReps[repertoire].append(extNeighSeq)

		return targetReps




"""
    seq_bit_impl
    ~~~~~~~~~~~~

    Contains classes that define alphabet types that can be used with the bit manipulation library.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

from seq_bit_lib import AbstractBitSeqManipulator
from genonets_utils import Utils


class BinaryBitSeqManipulator(AbstractBitSeqManipulator) :
	# No. of bits required to encode a single bit
	bitCodeLen = 1

	# Dictionary: Maps single bit letter to bit value 
	# (represented as integers here)
	letterToBitDict = {'0':0,  '1':1}

	# Dictionary: Maps bit values to bit letters
	bitToLetterDict = Utils.reverseDict(letterToBitDict)	


class ProteinBitSeqManipulator(AbstractBitSeqManipulator) :
	# No. of bits required to encode a single amino acid
	bitCodeLen = 5

	# Dictionary: Maps single letter amino acide codes to bit
	# values (represented as integers here)
	letterToBitDict = {'A':0,  'R':1,  'N':2,  'D':3,  'C':4,  \
					   'E':5,  'Q':6,  'G':7,  'H':8,  'I':9,  \
					   'L':10, 'K':11, 'M':12, 'F':13, 'P':14, \
					   'S':15, 'T':16, 'W':17, 'Y':18, 'V':19}

	# Dictionary: Maps bit values to single letter amino acide codes
	bitToLetterDict = Utils.reverseDict(letterToBitDict)	


# Abstract base class the provides common attributes for RNA
# DNA sequence manipulation as bits. Only derived concrete
# classes are supposed to be instantiated.
class AbstractXnaBitSeqManipulator(AbstractBitSeqManipulator) :
	# No. of bits required to encode a single nucleotide
	bitCodeLen = 2
	

# Concrete class for manipulation of arbitrary length RNA
# sequences as bits.
class RNABitSeqManipulator(AbstractXnaBitSeqManipulator) :
	# Dictionary that maps neuclotide letters to bit values 
	# (represented as integers here)
	letterToBitDict = {'A':0, 'U':1, 'C':2, 'G':3}
	
	# Dictionary that maps bit values to neuclotide letters
	bitToLetterDict = {0:'A', 1:'U', 2:'C', 3:'G'}


# Concrete class for manipulation of arbitrary length DNA
# sequences as bits.
class DNABitSeqManipulator(AbstractXnaBitSeqManipulator) :
	# Dictionary that maps neuclotide letters to bit values 
	# (represented as integers here)
	letterToBitDict = {'A':0, 'T':1, 'C':2, 'G':3}
	
	# Dictionary that maps bit values to neuclotide letters
	bitToLetterDict = {0:'A', 1:'T', 2:'C', 3:'G'}

	# Dictionary that maps each base to its complement
	baseToComplelentDict = {0: 1, 1: 0, 2: 3, 3: 2}

	# Computes and returns the bit representation of the reverse 
	# complement of the given sequence
	def getReverseComplement(self, sequence) :
		# Variable to hold the reverse complement
		revCompl = 0

		# Iterate over nucleotides in the given sequence 
		# in reverse order
		for i in reversed(range(self.seqLength)) :
			# Get nucleotide at index 'i'
			bitValue = self.getLetterAtIndex(sequence, i)

			# Add the complement base to the reverse complement
			revCompl |= self.baseToComplelentDict[bitValue]

			# If this is not the final nucleotide to be processed, 
			# i.e., the first nucleotide in the given sequence
			if i != 0 :
				# Left shift by 2 bits to make room for the next 
				# 2-bit sequence
				revCompl <<= 2

		return revCompl

	# Overrides the base implementation since for DNA, both
	# the sequence and its reverse complement have to be
	# considered.
	def areNeighbors(self, seq1, seq2) :
		# First check if seq1 and seq2 are 1-neighbors
		if AbstractXnaBitSeqManipulator.areNeighbors(self, seq1, seq2) :
			# If the two sequences are 1-neighbors, no need to 
			# process any further
			return True

		# Get the reverse complement of seq2
		revComplSeq2 = self.getReverseComplement(seq2)

		# Check if seq1 and reverse complement of seq2 are
		# 1-neighbors
		if AbstractXnaBitSeqManipulator.areNeighbors(self, seq1, revComplSeq2) :
			return True

		# At this point, neither possibility is true, so we return False.
		return False


# ------------------------------------------------------------
# Factory method
# ------------------------------------------------------------

# TODO: Could probably do something similar to AnalysisHandler here, i.e.,
#		use a more elegant method for determining the BitSeqManipulator
#		corresponding to the given molecule type. E.g., using a dict,
#		{moleculeType : constructor} ...
class BitManipFactory :
	# Supported molecule types
	moleculeTypes = ["RNA", "DNA", "Protein", "Binary"]

	# Returns the list of supported molecule types
	@staticmethod
	def getMoleculeTypes() :
		return BitManipFactory.moleculeTypes

	# Based on the given molecule type, returns the corresponding
	# bit seq manipulator object.
	@staticmethod
	def getBitSeqManip(moleculeType, seqLength, useIndels) :
		if moleculeType == "RNA" :
			return RNABitSeqManipulator(seqLength, useIndels)
		elif moleculeType == "DNA" :
			return DNABitSeqManipulator(seqLength, useIndels)
		elif moleculeType == "Protein" :
			return ProteinBitSeqManipulator(seqLength, useIndels)
		elif moleculeType == "Binary" :
			return BinaryBitSeqManipulator(seqLength, useIndels)
		else :
			print("Unsupported moleculeType: " + str(moleculeType))
			print("Exiting program ...")

			exit()




"""
    structure_functions
    ~~~~~~~~~~~~~~~~~~~

    Encapsulates functions for structural analyses.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

import igraph


class StructureAnalyzer :
	# Constructor
	def __init__(self, network, netBuilder) :
		# Reference to the network on which to perform this
		# analysis
		self.network = network

		# Get a reference to the NetworkBuilder object
		self.netBuilder = netBuilder

		# Reference to giant
		self.giant = self.netBuilder.getGiantComponent(self.network)

	# ----------------------------------------------------------------
	#	Network level properties
	# ----------------------------------------------------------------

	def getComponentSizes(self) :
		return self.netBuilder.getComponents(self.network)

	def getNumComponents(self) :
		return len(self.netBuilder.getComponents(self.network))

	def getDominantSize(self) :
		return self.giant.vcount()

	def getPercentDominantSize(self) :
		return (float(self.giant.vcount()) / float(self.network.vcount()))

	def getEdgeDensity(self) :
		return self.giant.density()

	def getAvgClstrCoeff(self) :
		return self.giant.transitivity_avglocal_undirected()

	def getAssortativity(self) :
		return self.giant.assortativity_degree()

	def getDiameter(self) :
		return self.giant.diameter()

	def getDiameterPath(self) :
		# Get network diameter
		diameter = self.getDiameter()

		# Get the vertex sequence
		vSeq = self.giant.vs
		
		# Calculate shortest paths between each pair of vertices in the
		# network until one is found of size 'diameter'.
		for i in range(len(vSeq) - 1) :
			for j in range(i+1, len(vSeq)) :
				# Calculate shortest path between vertx 'i' and vertex 'j'
				sPath = self.giant.get_shortest_paths(vSeq[i], vSeq[j], 
					mode=igraph.OUT, output="vpath")

				# Return the path is the No. of vertices in the path
				# is diameter + 1, i.e, the No. of edges = diameter.
				if len(sPath[0]) == (diameter + 1) :
					return sPath[0]

		return []

	# ----------------------------------------------------------------
	#	Vertex level properties
	# ----------------------------------------------------------------

	def getCoreness(self) :
		return self.giant.coreness()

	# Returns the clustering coefficients for all sequences in the giant
	# component.
	def getClusteringCoefficients(self) :
		# Get all sequences
		sequences = self.giant.vs

		# Get the list of clustering coefficients for all vertices
		# in giant. The list of sequences is passed as an argument
		# just to make sure that the order of coefficients in the
		# list corresponds to the order of passed sequences.
		clstrCoeffs = self.giant.transitivity_local_undirected(sequences)

		return clstrCoeffs




"""
    evolvability_functions
    ~~~~~~~~~~~~~~~~~~~~~~

    Contains functions used for epistasis computation.

    :author: Fahad Khalid
    :license: MIT, see LICENSE for more details.
"""

from genonets_constants import EpistasisConstants as epi


class EpistasisAnalyzer :
	# Constructor
	def __init__(self, network, netUtils, seqToEscrDict, delta, bitManip) :
		# Reference to the network on which to perform this
		# analysis
		self.network = network

		# Get reference to the NetworkUtils object
		self.netUtils = netUtils

		# Get reference to dict: key=sequence, value=escore
		self.seqToEscrDict = seqToEscrDict

		# Keep a copy of the delta value
		self.delta = delta

		# Refernce to the bit manipulator object
		self.bitManip = bitManip

		# Dict {sequence : vertex ID}
		self.seqToVidDict = self.buildSeqToVidDict()

		# To store the computed squares
		self.squares = None

		# To store the computed epistasis class per square
		self.sqrEpi = None

	# Computes epistasis for the given network
	def getEpiAll(self) :
		# Dictionary to keep count of the number of squares that
		# show each class of epistasis.
		epistasis = {epi.MAGNITUDE : 0, epi.SIGN : 0, \
					 epi.RECIPROCAL_SIGN : 0, epi.NO_EPISTASIS : 0}

		# Get all squares
		squares = self.getSquares()

		# List of epistasis class corresponding to each square
		self.sqrEpi = []

		# For each square
		for square in squares :
			# Determine to which class of epistasis this square belongs
			epiClass = self.getEpistasis(square)

			# Store epistasis type for this square
			self.sqrEpi.append(epiClass)

			# Increment the count for the epistasis class to which the
			# square belongs
			epistasis[epiClass] += 1

		# There's no need to keep count of no epistasis
		del epistasis[epi.NO_EPISTASIS]

		# Convert counts into ratios
		for epiClass in epistasis.keys() :
			try :
				epistasis[epiClass] = float(epistasis[epiClass]) / float(len(squares))
			except ZeroDivisionError :
				epistasis[epiClass] = 0

		return epistasis

	def getEpistasis(self, square) :
		epsilon = 0
		epiClass = epi.NO_EPISTASIS

		# Get e-scores for all corners of the square
		esrc_AB = self.seqToEscrDict[square[3]]
		esrc_ab = self.seqToEscrDict[square[0]]
		esrc_Ab = self.seqToEscrDict[square[2]]
		esrc_aB = self.seqToEscrDict[square[1]]

		# Calculate the magnitude of epistasis
		epsilon = esrc_AB + esrc_ab - esrc_Ab - esrc_aB

		# If epsilon is greater than or equal to the noise threshold,
		# there is epistasis
		if abs(epsilon) >= self.delta :
			# Classify the type of epistasis
			epiClass = self.getEpiClass(esrc_ab, esrc_aB, esrc_Ab, esrc_AB)

			# Check if we have a case where all mutational effects are 0
			if epiClass == epi.NO_EPISTASIS :
				# No epistasis
				epsilon = 0
		else :
			# No epistasis
			epsilon = 0

		return epiClass

	def getEpiClass(self, esrc_ab, esrc_aB, esrc_Ab, esrc_AB) :
		epiClass = 0

		# Calculate mutational effects
		dE_ab_Ab = esrc_Ab - esrc_ab
		dE_aB_AB = esrc_AB - esrc_aB
		dE_ab_aB = esrc_aB - esrc_ab
		dE_Ab_AB = esrc_AB - esrc_Ab

		# A mutational effect should only be considered if it is larger
		# than delta
		dE_ab_Ab = dE_ab_Ab if abs(dE_ab_Ab) >= self.delta else 0
		dE_aB_AB = dE_aB_AB if abs(dE_aB_AB) >= self.delta else 0
		dE_ab_aB = dE_ab_aB if abs(dE_ab_aB) >= self.delta else 0
		dE_Ab_AB = dE_Ab_AB if abs(dE_Ab_AB) >= self.delta else 0

		# If all mutational effects are 0, there is no epistasis
		if dE_ab_Ab == 0 and dE_aB_AB == 0 and dE_ab_aB == 0 and dE_Ab_AB == 0 :
			return epi.NO_EPISTASIS

		# Conditions used to determine epistasis type
		condition1 = abs(dE_ab_Ab + dE_aB_AB) == abs(dE_ab_Ab) + abs(dE_aB_AB)
		condition2 = abs(dE_ab_aB + dE_Ab_AB) < abs(dE_ab_aB) + abs(dE_Ab_AB)

		# Evaluate the the conditions to determine epistasis type
		if condition1 and not condition2 :
			# if condition 1 holds but not condition 2, there is 
			# magnitude epistasis
			epiClass = epi.MAGNITUDE
		elif condition2 and not condition1 :
			# if condition 2 holds but not condition 1, there is 
			# reciprocal sign epistasis
			epiClass = epi.RECIPROCAL_SIGN
		else :
			# Otherwise, we have sign epistasis
			epiClass = epi.SIGN

		return epiClass

	# ----------------------------
	# Squares
	# ----------------------------

	def buildSeqToVidDict(self) :
		# Reference to the list of sequences
		sequences = self.network.vs["sequences"]

		return { sequences[i] : i for i in range(len(sequences)) }

	# Return all squares found in the genotype network
	# TODO: Look into performance optimization ...
	def getSquares(self, recompute=False) :
		# If squares computation has been done already, and the 
		# caller has not explicitly asked for re-running the 
		# algorithms,
		if self.squares and not recompute :
			# Return the pre-computed results
			return self.squares

		# List of squares to be populated
		squares = []

		# Set of unique squares (permutation agonistic) with sequences
		# in bit format
		bitSqrs = set()

		# Get the list of all sequences
		sequences = self.network.vs["sequences"]

		# For each sequence in the network
		for sequence in sequences :
			# Get all 1-neighbor sequences
			neighbors = [self.network.vs[vid]["sequences"] \
							for vid in \
							self.network.neighbors( \
								self.netUtils.getVertex(sequence, self.network))]

			# If the number of neighbors is less than two, there's no point
			# in continuing any further
			if len(neighbors) < 2 :
				# Move on to the next sequence
				continue

			# Construct all possible pairs of neighbors, where symmetric pairs
			# are considered only once. Also, pairs that neighbor each other
			# are not considered.
			pairs = [ (neighbors[i], neighbors[j]) \
						for i in range(len(neighbors) - 1) \
						for j in range(i + 1, len(neighbors)) \
						if not self.netUtils.areConnected( \
							neighbors[i], neighbors[j]) ]

			# For each pair of neighbors
			for pair in pairs :
				# Get neighbors common to both sequences in the pair, but
				# without the sequence itself
				commonNeighs = self.getCommonNeighbors(pair, sequence)

				# For each common neighbor
				for node in commonNeighs :
					# For a node that is an immediate neighbor of the parent,
					# it does not make sense to compute epistasis, since a
					# single mutation from parent is sufficient.
					if node not in neighbors :
						# Construct the square
						square = [sequence, pair[0], pair[1], node]
						# Square as a set of bitseqs
						bitSqr = frozenset( [self.bitManip.seqToBits(sequence),
									   self.bitManip.seqToBits(pair[0]), 
									   self.bitManip.seqToBits(pair[1]),
									   self.bitManip.seqToBits(node)] )

						# If the square has not already been found.
						# Note: This is a set operation, which means due to 
						# the lack or ordering, all permutations will be tested 
						# with just this one condition.
						if not bitSqr in bitSqrs :
							squares.append(square)
							bitSqrs.add(bitSqr)

		# Keep a reference for re-use
		self.squares = squares

		return squares

	# Returns a list of neighbors common to both elements in the pair. The
	# parent is not considered as a common neighbor.
	def getCommonNeighbors(self, pair, parent) :
		# Get neighbors for the first sequence in the pair
		neighbors1 = [self.network.vs[vid]["sequences"] for vid in \
						self.network.neighbors(self.seqToVidDict[pair[0]])]

		# Get neighbors for the second sequence in the pair
		neighbors2 = [self.network.vs[vid]["sequences"] for vid in \
						self.network.neighbors(self.seqToVidDict[pair[1]])]

		# Get a list of common neighbors
		commonNeighbors = list(set(neighbors1) & set(neighbors2))

		# Remove the parent
		commonNeighbors.remove(parent)

		return commonNeighbors

	# Return a dict - {verterx id : [square ids in which this vertex appears]}
	def getVertexToSquaresDict(self) :
		# Make sure there is at least one square
		if not self.squares or len(self.squares) < 1 :
			return {}, []

		# Get the list of all sequences
		sequences = self.network.vs["sequences"]

		# Initialize the dict: Keys = vertex ids corresponding to sequence indices
		# in the list of sequences; Values = empty lists to be populated later with
		# square ids
		vtxToSqrs = { vId : [] for vId in range(len(sequences)) }

		# List of squares with vertex ids instead of sequences as square elements
		squares = []

		# For each square,
		for sqrIndx in range(len(self.squares)) :
			sqr = []
			# For each sequence in the square,
			for sequence in self.squares[sqrIndx] :
				# Get the vertex id corresponding to the sequence
				seqIndx = sequences.index(sequence)				

				# Add the square Id as value to the vertex Id
				vtxToSqrs[seqIndx].append(sqrIndx)

				# Add vetex id to sqr
				sqr.append(seqIndx)

			# Append the square to the list of squares
			squares.append(sqr)

		return vtxToSqrs, squares

	def getSqrEpi(self) :
		return self.sqrEpi


#!/usr/bin/env python

from setuptools import setup

setup(	name='genonets',
		version='1.0.0',
		description='Framework for creating and analyzing genotype networks from data.',
		author='Fahad Khalid',
		author_email='fahad.khalid@ieu.uzh.ch',
		url='https://github.com/fkhalid/genonets',
		license='MIT',
		classifiers=[
			'Development Status :: 4 - Beta',
			'Intended Audience :: Science/Research',
			'License :: OSI Approved :: MIT License',
			'Programming Language :: Python :: 2.7',
			'Operating System :: OS Independent',
			'Natural Language :: English',
			'Topic :: Scientific/Engineering :: Bio-Informatics',			
		],
		keywords='genonets genotype phenotype network',
		packages=['genonets'],
		package_data={'genonets': ['sample/*.py', 'sample/*.txt', 'sample/data/*.txt']},
		install_requires=['python-igraph>=0.6', 'numpy>=1.8.2']
	)
